"""Migrate existing samples to most current slybot version.

Convert tagid to selector
    1. Add tagids
    2. Load page in lxml
    3. for each annotation
    4.     find element by tagid
    5.     get unique selector for tagid
    6.     save unique selector

Handle variants
    1. Find matching variants
    2. Find selector for each matching variant
    3. Find common parent
    4. Find repeated parent
    5. Assign to parent field called variant

Handle generated annotations
    1. Find element referred to by annotation
    2. Find text for element
    3. Find text for annotation
    3. Remove extra annotation metadata
"""
import copy
import json
import re

import slybot

from collections import defaultdict
from itertools import chain, groupby
from operator import itemgetter
from random import Random
from urllib import unquote
from uuid import uuid4

from lxml.etree import _Element, Comment
from scrapy import Selector

from slybot.utils import add_tagids
SLYBOT_VERSION = slybot.__version__
IGNORE_ATTRIBUTES = ['data-scrapy-ignore', 'data-scrapy-ignore-beneath']
_ID_RE = re.compile('([0-9a-f]{4}-){2}[0-9a-f]{4}')


def short_guid():
    return '-'.join(str(uuid4()).split('-')[1:4])


def gen_id(disallow=None):
    if disallow is not None:
        disallow = set(disallow)
    else:
        disallow = []
    _id = short_guid()
    while _id in disallow:
        _id = short_guid()
    return _id


def gen_predictable_id(id1, id2, disallow=None):
    if disallow is not None:
        disallow = set(disallow)
    else:
        disallow = []
    seed = id_to_int(id1) ^ id_to_int(id2)
    generator = Random(seed)

    def format_id():
        id_ = '{:0>12x}'.format(generator.randint(0, 2**47))
        return '-'.join((id_[:4], id_[4:-4], id_[-4:]))
    id_ = format_id()
    while id_ in disallow:
        id_ = format_id()
    return id_


def id_to_int(id_):
    try:
        return int(str(id_).replace('-', ''), 16)
    except ValueError:
        seed = 1
        for i, ch in enumerate(id_):
            seed *= ord(ch) + i
        return seed


def repair_ids(sample):
    annotations = sample['plugins']['annotations-plugin']['extracts']
    aids, cids = defaultdict(list), {}
    for a in annotations:
        if a.get('item_container'):
            cids[a['id']] = a
        if a.get('container_id'):
            aids[a['container_id']].append(a)
    full_ids = PartialKeyDict()
    for id_ in aids:
        try:
            full_ids[id_] = id_
        except ValueError:
            pass
    for id_, affected in aids.items():
        if not id_ or len(id_) in (14, 21):
            continue
        full_id = full_ids[id_]
        if id_ != full_id:
            for a in affected:
                a['container_id'] = full_id
            if id_ in cids:
                cids[id_]['id'] = full_id
    return sample


def port_sample(sample, schemas=None, extractors=None):
    """Convert slybot samples made before slybot 0.13 to new format."""
    if schemas is None:
        schemas = {}
    if extractors is None:
        extractors = {}
    container_id = gen_predictable_id(sample.get('id', 1), sample['page_id'])
    default_annotations = [_create_container('body', container_id)]
    if not sample.get('annotated_body') and not sample.get('plugins'):
        sample['plugins'] = {
            'annotations-plugin': {
                'extracts': default_annotations
            }
        }
        return sample
    if not sample.get('plugins'):
        sample['plugins'] = load_annotations(sample.get('annotated_body', u''))
    else:
        repair_ids(sample)
    sample.pop('annotated_body', None)

    # Group annotations by type
    annotations = sample['plugins']['annotations-plugin']['extracts']
    try:
        sel = Selector(text=add_tagids(sample['original_body']))
    except KeyError:
        annotated = sample.get('annotated_body', u'')
        sample['original_body'] = annotated
        try:
            tagged = add_tagids(annotated)
        except KeyError:
            tagged = u''
        sel = Selector(text=tagged)
    annotations = port_standard(annotations, sel, sample)
    standard_annos, generated_annos, variant_annos = [], [], []
    for a in annotations:
        if a.get('generated'):
            generated_annos.append(a)
        elif a.get('variants', 0) > 0:
            variant_annos.append(a)
        else:
            standard_annos.append(a)
    if not annotations:
        sample['plugins'] = {
            'annotations-plugin': {
                'extracts': default_annotations
            }
        }
        return sample
    new_annotations = []
    a = find_element(annotations[0], sel)
    for b in annotations[1:]:
        b = find_element(b, sel)
        a = find_common_parent(a, b)
    parent = a.getparent()
    container = _create_container(
        a if parent is None else parent, container_id, selector=sel)
    new_annotations.append(container)
    for a in standard_annos:
        a.pop('variant', None)
    new_annotations.extend(standard_annos)
    new_annotations.extend(port_generated(generated_annos, sel))
    new_annotations.extend(port_variants(variant_annos, sel))
    for a in new_annotations:
        if not (a.get('item_container') and a.get('container_id')):
            a['container_id'] = container_id
        a.pop('tagid', None) or a.pop('data-tagid', None)
    # Update annotations
    sample['plugins']['annotations-plugin']['extracts'] = new_annotations
    sample['version'] = SLYBOT_VERSION
    schema_id, schemas = guess_schema(sample, schemas)
    container['schema_id'] = schema_id
    return sample, schemas


def find_element(tagid, sel):
    """Find an element by its tagid."""
    if isinstance(tagid, _Element):
        return tagid
    if isinstance(tagid, dict):
        tagid = tagid.get('tagid')
    elements = sel.xpath('//*[@data-tagid="%s"]' % tagid)
    if elements:
        return elements[0].root


def css_escape(s):
    """from http://mathiasbynens.be/notes/css-escapes"""
    s = re.sub(r'(^-$|[ !"#\$%&\'()*+,./:;<=>?@\[\\\]^`{|}~])', r'\\\1', s)
    s = re.sub(r'^(-?)(\d)', lambda m: '%s\%x ' % (m.group(1), ord(m.group(2))), s)
    s = re.sub(r'([\t\n\v\f\r])', lambda m: '\%x ' % ord(m.group(1)), s)
    return s


def find_css_selector(elem, sel, depth=0, previous_tbody=False):
    """Find a unique selector for an element.

    Adapted from mozilla findCssSelector in css-logic.js
    http://lxr.mozilla.org/mozilla-release/source/toolkit/devtools/styleinspector/css-logic.js
    """
    def children_index(elem):
        parent = elem.getparent()
        if parent is not None:
            children = filter(lambda x: x.tag is not Comment,
                              parent.getchildren())
            index = children.index(elem) + 1
        else:
            index = 0
        return index

    def build_table_selector(elem):
        parent = find_css_selector(elem.getparent(), sel, depth + 1)
        join = '' if previous_tbody else ' >'
        selector = '%s%s %s:nth-child(%s)' % (
            parent, join, tag_name, children_index(elem)
        )
        e = sel.css(selector)
        if not e:
            join = '' if previous_tbody or tag_name == 'tr' else ' >'
            selector = '%s%s %s:nth-child(%s)' % (
                parent, join, tag_name, children_index(elem))
        return selector

    elem_id = elem.attrib.get('id')
    if elem_id:
        id_selector = '#%s' % css_escape(elem_id)
        if len(sel.css(id_selector)) == 1 and depth <= 1:
            return id_selector

    # Inherently unique by tag name
    tag_name = elem.tag
    for tag in ('html', 'head', 'body'):
        if tag_name == tag:
            return tag

    # We might be able to find a unique class name
    classes = elem.get('class', '').split()
    if classes:
        for class_name in classes:
            selector = '.%s' % css_escape(class_name)
            matches = sel.css(selector)
            if len(matches) == 1:
                return selector
            # Maybe it's unique with a tag name?
            selector = tag_name + selector
            matches = sel.css(selector)
            if len(matches) == 1:
                return selector
            # Maybe it's unique using a tag name and nth-child
            selector = '%s:nth-child(%s)' % (selector, children_index(elem))
            matches = sel.css(selector)
            if len(matches) == 1:
                return selector

    # Not unique enough yet.  As long as it's not a child of the document,
    # continue recursing up until it is unique enough.
    if elem.getparent() is not None:
        if tag_name in ('thead', 'tbody'):
            selector = find_css_selector(elem.getparent(), sel, depth + 1,
                                         True)
        else:
            selector = build_table_selector(elem)
    return selector


def find_common_parent(a, b):
    """Find a common parent for 2 elements."""
    a_parents = list(a.iterancestors())
    b_parents = list(b.iterancestors())
    a_parents_set = set(a_parents)
    b_parents_set = set(b_parents)
    if a == b:
        return a
    if b in a_parents_set:
        return b
    if a in b_parents_set:
        return a
    if len(a_parents) < len(b_parents):
        for elem in a_parents:
            if elem == b or elem in b_parents_set:
                return elem
    else:
        for elem in b_parents:
            if elem == a or elem in a_parents_set:
                return elem


def port_variants(variant_annotations, sel):
    """Port variant annotations to the MIE annotations."""
    # Group variants
    grouper = itemgetter('variants')
    grouped_variants = [(vid, list(grp))
                        for vid, grp in groupby(variant_annotations, grouper)]
    adjacent_variants = set()
    # Eliminate non adjacent variants
    for variant_id, _ in grouped_variants:
        if not variant_id:
            continue
        if variant_id in adjacent_variants:
            adjacent_variants.remove(variant_id)
        else:
            adjacent_variants.add(variant_id)
    # Remove non sequential variants
    grouped_variants = [(vid, variant) for vid, variant in grouped_variants
                        if variant_id in adjacent_variants and variant]
    annotations = []
    for first, last in zip(*[iter(grouped_variants)] * 2):
        siblings = 0
        repeated_container = _get_parent(first, sel)
        last_shared_parent = _get_parent(last, sel)
        container = find_common_parent(repeated_container, last_shared_parent)
        if repeated_container == container:
            repeated_container, siblings = _get_parent_and_siblings(first,
                                                                    container,
                                                                    sel)
        container_id = gen_predictable_id(first['id'], last['id'])
        for annotation in first:
            annotation['container_id'] = container_id
            del annotation['variant']
        annotations.extend(first)
        annotations.append(_create_container(container, container_id,
                                             field='variants', selector=sel))
        annotations.append(_create_container(repeated_container, container_id,
                                             repeated=True, siblings=siblings,
                                             selector=sel))
    return variant_annotations


def _get_parent(annotations, sel):
    parent = annotations[0]
    parent_element = find_element(parent, sel)
    if len(annotations) == 1:
        return parent_element.getparent()
    for annotation in annotations[1:]:
        annotation_element = find_element(annotation, sel)
        parent_element = find_common_parent(parent_element, annotation_element)
    return parent_element


def _get_parent_and_siblings(annotations, upto, sel):
    # Find deepest parents at same depth from shared parent which are siblings
    highest = _get_highest(annotations, upto, sel)
    highest_reverse = _get_highest(list(reversed(annotations)), upto, sel)
    parent_siblings = [set(highest_reverse.getparent().getchildren())]
    for parent in highest_reverse.iterancestors():
        if parent == upto:
            break
        parent_siblings.append(set(parent.getparent().getchildren()))
    for element in chain((highest,), highest.iterancestors()):
        for siblings_set in parent_siblings:
            if element in siblings_set:
                highest = element
                break
    # Find distance between those siblings
    siblings = highest.getparent().getchildren()
    siblings_set = set(siblings)
    max_sibling = 0
    for annotation in annotations:
        parents = set(find_element(annotation, sel).iterancestors())
        intersection = parents & siblings_set
        if not intersection:
            continue
        try:
            sibling = siblings.index(list(intersection)[0])
        except ValueError:
            continue
        if sibling > max_sibling:
            max_sibling = sibling
    return highest, sibling


def _get_highest(annotations, upto, sel):
    highest = find_element(annotations[0], sel)
    for annotation in annotations[1:]:
        element = find_element(annotation, sel)
        next = find_common_parent(highest, element)
        if next == upto:
            break
        highest = next
    return highest


def _create_container(element, container_id, repeated=False, siblings=0,
                      field=None, selector=None):
    if isinstance(element, str):
        s = element
    else:
        s = find_css_selector(element, selector)
    data = {
        'id': '%s%s' % (container_id, '#parent' if repeated else ''),
        'accept_selectors': [s],
        'reject_selectors': [],
        'selector': s,
        'item_container': True,
        'repeated': repeated,
        'required': [],
        'siblings': siblings,
        'annotations': {'#portia-content': '#dummy'},
        'text-content': '#portia-content',
    }
    if repeated:
        data['container_id'] = container_id
    if hasattr(element, 'attrib') and 'data-tagid' in element.attrib:
        data['data-tagid'] = element.attrib['data-tagid']
    if field:
        data['field'] = field
    return data


def _add_annotation_data(annotation, sample, extractors):
    if 'data' in annotation:
        return annotation
    annotations = sample['plugins']['annotations-plugin']['extracts']
    existing_ids = {i for a in annotations for i in a.get('data', [])}
    annotation['data'] = {}
    for attribute, field in annotation.get('annotations', {}).items():
        if field == '#dummy':
            continue
        _id = gen_predictable_id(
            sample['page_id'], attribute + field or '', disallow=existing_ids)
        extractors = [e for e in sample.get('extractors', {}).get(field, [])
                      if e in extractors]
        annotation['data'][_id] = {
            'attribute': attribute,
            'field': field,
            'required': field in annotation.get('required', []),
            'extractors': extractors
        }
    return annotation


def port_generated(generated_annotations, sel):
    """Port generated annotations to annotations with custom extractors."""
    for annotation in generated_annotations:
        # Find pre and post text
        _slice = annotation.get('slice')
        if not annotation.get('tagid') or not _slice or len(_slice) != 2:
            continue
        element = find_element(annotation, sel)
        if element is None:
            continue
        pre_text, post_text = '', ''
        if annotation.get('insert_after'):
            selector = find_css_selector(element.getparent(), sel)
            if not selector:
                continue
            annotation['accept_selectors'] = [selector]
            annotation['selector'] = selector
            siblings = list(element.getparent().iterchildren())
            start = siblings.index(element)

            def process_siblings(others, reverse=False):
                text = []
                order = ('tail', 'text') if reverse else ('text', 'tail')
                for elem in others:
                    for t in order:
                        text.append(getattr(elem, t) or ' ')
                    if any(t.strip() for t in text):
                        break
                return ''.join(text)
            pre_text = process_siblings(
                siblings[start - 1: 0: -1] + [siblings[0]], True)
            post_text = process_siblings(siblings[start + 1:], 1)
            text = element.tail
        else:
            text = element.text
        text = (text or '').strip()
        data = annotation['data']
        contentf = annotation.get('text-content', 'content')
        for anno in (a for a in data.values() if a['attribute'] == contentf):
            pre_text = (pre_text + text[0:_slice[0]]).strip()
            post_text = (text[_slice[1]:] + post_text).strip()
            if anno and pre_text or post_text:
                anno['pre_text'] = pre_text
                anno['post_text'] = post_text
        # Create new text region field
        annotation.pop('variant', None)
        del annotation['slice']
    return generated_annotations


def port_standard(standard_annotations, sel, sample, extractors):
    """Add accept selectors for existing annotations."""
    new_annotations = []
    for annotation in standard_annotations:
        if not annotation.get('tagid'):
            continue
        element = find_element(annotation, sel)
        if element is None:
            continue
        selector = find_css_selector(element, sel)
        if not selector:
            continue
        annotation['accept_selectors'] = [selector]
        annotation['selector'] = selector
        annotation['reject_selectors'] = []
        annotation = _add_annotation_data(annotation, sample, extractors)
        for _id, data in annotation.get('data', {}).items():
            a = copy.deepcopy(annotation)
            a['id'] = gen_predictable_id(_id, a['id'])
            a['data'] = {
                gen_predictable_id(a['id'], 1): data
            }
            new_annotations.append(a)
    return new_annotations


def load_annotations(body):
    """Create slybot annotations from annotated html."""
    if not body:
        return {'annotations-plugin': {'extracts': []}}
    sel = Selector(text=add_tagids(body))
    existing_ids = set()
    annotations = []
    for elem in sel.xpath('//*[@data-scrapy-annotate]'):
        attributes = elem.root.attrib
        annotation = json.loads(unquote(attributes['data-scrapy-annotate']))
        if (isinstance(elem.root, _Element) and
                elem.root.tag.lower() == 'ins'):
            annotation.update(find_generated_annotation(elem))
        else:
            annotation['tagid'] = attributes.get('data-tagid')
        if 'id' not in annotation:
            annotation['id'] = gen_id(disallow=existing_ids)
        existing_ids.add(annotation['id'])
        annotations.append(annotation)
    for elem in sel.xpath('//*[@%s]' % '|@'.join(IGNORE_ATTRIBUTES)):
        attributes = elem.root.attrib
        for attribute in IGNORE_ATTRIBUTES:
            if attribute in attributes:
                break
        ignore = {attribute[len('data-scrapy-'):]: True}
        if 'id' not in ignore:
            ignore['id'] = gen_id(disallow=existing_ids)
        existing_ids.add(ignore['id'])
        annotations.append(ignore)
    return {'annotations-plugin': {'extracts': annotations}}


def find_generated_annotation(elem):
    """Find annotation information for generated element."""
    elem = elem.root
    previous = elem.getprevious()
    insert_after = True
    nodes = []
    if previous is None:
        previous = elem.getparent()
        nodes = [previous] + previous.getchildren()
        insert_after = False
    else:
        while (previous is not None and isinstance(previous, _Element) and
                previous.tag.lower() == 'ins'):
            previous = previous.getprevious()
        if previous is None:
            previous = elem.getparent()
            insert_after = False
            node = previous.getchildren()[0]
        else:
            node = previous
        while node is not None:
            nodes.append(node)
            node = node.getnext()
            if (node is None or
                    isinstance(node, _Element) and node.tag.lower() == 'ins'):
                nodes.append(node)
                break
    annotation = {
        'tagid': previous.attrib.get('data-tagid'),
        'generated': True,
        'insert_after': insert_after
    }
    last_node_ins = False
    start = 0
    # Calculate the length and start position of the slice ignoring the ins
    # tag and with leading whitespace removed
    for node in nodes:
        if (node is not None and isinstance(node, _Element) and
                node.tag.lower() == 'ins'):
            last_node_ins = True
            if node == elem:
                annotation['slice'] = start, start + len(node.text)
            else:
                start += len(node.text)
        else:
            text = node.tail if annotation['insert_after'] else node.text
            text = '' if text is None else text
            if not last_node_ins:
                text = text.lstrip()
            start += len(text)
            last_node_ins = False
    return annotation


def guess_schema(sample, schemas):
    schema_id = _guess_schema_id(sample, schemas)
    annotations = sample['plugins']['annotations-plugin']['extracts']
    if schema_id not in schemas:
        schemas[schema_id] = create_schema(schemas, annotations)
    else:
        schemas[schema_id] = add_fields(schemas[schema_id], annotations)
    return schema_id, schemas


def _guess_schema_id(sample, schemas):
    # If project has no schemas just return default value
    if not schemas:
        return 'default'

    # Check if scrapes exists
    schema_id = sample.get('scrapes')
    if schema_id and schema_id in schemas:
        return schema_id

    # Check if a schema is explicitly mentioned
    annotations = sample['plugins']['annotations-plugin']['extracts']
    annotations_with_schemas = [a for a in annotations if 'schema_id' in a]
    if annotations_with_schemas:
        parent = sorted(annotations_with_schemas, key=container_id_key)[0]
        return parent['schema_id']

    # Look for a schema with matching fields
    fields = set()
    for annotation in annotations:
        try:
            fields.add(list(annotation['data'].values())[0]['field'])
        except KeyError:
            pass
    scores = {}
    fields = set()
    for schema_id, schema in schemas.items():
        scores[schema_id] = len(set(schema.get('fields')) & fields)
    if any(score > 0 for score in scores.values()):
        return max(scores.items(), key=itemgetter(1))[0]

    # Use default if it is available
    if 'default' in schemas:
        return 'default'

    # Use the schema with the most fields
    schemas = sorted(
        schemas.items(), key=lambda (k, v): len(v.get('fields', {})))
    return schemas[-1][0]


def add_fields(schema, annotations):
    field_ids = set(schema['fields'].keys())
    schema['fields'].update(_create_fields(annotations, field_ids))
    return schema


def create_schema(schemas, annotations):
    fields = _create_fields(annotations, ())
    schema = {'name': 'schema_%s' % (len(schemas) + 1),
              'fields': fields}
    return schema


def _create_fields(annotations, field_ids):
    field_ids = set(filter(bool, field_ids))
    fields, field_number = {}, len(field_ids) + 1
    for a in sorted(annotations, key=lambda a: a['id']):
        if 'data' not in a:
            continue
        for aid, attribute in a['data'].items():
            field_name = attribute['field']
            if field_name in field_ids:
                # Skip existing fields
                continue
            if field_name is None:
                field_name = gen_predictable_id(a['id'], aid)
                a['data'][aid]['field'] = field_name
            field, field_number = _field(field_name, field_number)
            fields[field_name] = field
            field_ids.add(field_name)
    return fields


def _field(field_id, num_fields):
    field_name = field_id
    if _ID_RE.match(field_id):
        field_name = 'field%d' % num_fields
        num_fields += 1
    field = {'name': field_name, 'vary': False,
             'required': False, 'type': 'text'}
    return field, num_fields


def container_id_key(annotation):
    container_id = annotation.get('', 'z' * 19)
    if container_id is None:
        return ''
    return container_id


class PartialKeyDict(dict):
    def __getitem__(self, key):
        try:
            return super(PartialKeyDict, self).__getitem__(key)
        except KeyError:
            removed_parent, key = self._remove_parent(key)
            if len(key) == 13:
                full_key = self._find_key(key)
                if full_key:
                    if removed_parent:
                        key, full_key = self._add_parent(key, full_key)
                    self[key] = full_key
                    return full_key
            raise

    def __setitem__(self, key, value):
        removed_parent, key = self._remove_parent(key)
        if key and len(key) not in (14, 21):
            full_key = self._find_key(key)
        else:
            full_key = key
        if len(full_key) != 14:
            raise ValueError('ids must be of length 14')
        if removed_parent:
            key, full_key = self._add_parent(key, full_key)
        if full_key:
            super(PartialKeyDict, self).__setitem__(key, full_key)

    def _find_key(self, key):
        for k in self:
            if k.startswith(key):
                return k
        return None

    def _remove_parent(self, key):
        if key.endswith('#parent'):
            return True, key[:-len('#parent')]
        return False, key

    def _add_parent(self, key, full_key):
        parent_fmt = u'{}#parent'.format
        return parent_fmt(key), parent_fmt(full_key)



from itertools import chain


class GeneratedUrl(object):
    def __init__(self, spec):
        self.key = spec
        self.spec = spec
        self.generator_value = spec
        self.generator_type = 'generated_urls'

    @property
    def allowed_domains(self):
        return [self.spec['template']]

    @property
    def normalized(self):
        return {
            'url': self.normalized_url,
            'type': 'generated',
            'fragments': self.normalized_fragments,
        }

    @property
    def normalized_url(self):
        paths = [normalize_url_path(path) for path in self.spec['paths']]
        base_url = self.spec['template'].format(*paths)

        query_params = [normalize_url_query_param(path, is_first=(i == 0))
                        for i, path in enumerate(self._query_params)]
        return base_url + ''.join(query_params)

    @property
    def normalized_fragments(self):
        fixed, path = self._fixed_fragments, self._path_fragments
        fragments = list(zip(fixed, path))

        # Missed last fixed fragment when using zip
        if len(fixed) == len(path) + 1:
            fragments.append([fixed[-1]])

        return list(chain.from_iterable(fragments)) + self._query_fragments

    @property
    def _path_fragments(self):
        return [normalize_path(path) for path in self.spec['paths']]

    @property
    def _fixed_fragments(self):
        return [{'type': 'fixed', 'value': fragment} for fragment in
                self.spec['template'].split('{}') if fragment]

    @property
    def _query_fragments(self):
        if len(self._query_params) == 0:
            return []

        fragments = [normalize_query_param(query_param, is_first=(i == 0))
                     for i, query_param in enumerate(self._query_params)]
        return list(chain(*fragments))

    @property
    def _query_params(self):
        params = self.spec['params']
        template_params = self.spec['params_template']

        query_params = []
        param_names = {p['name'] for p in params}
        for query, value in template_params:
            if query not in param_names:
                query_params.append({
                    'name': query,
                    'type': 'default',
                    'values': [value]
                })
        return query_params + params


def normalize_path(path):
    if path['type'] == 'options':
        return {
            'type': 'list',
            'value': ' '.join(path['values']),
        }
    if path['type'] == 'default':
        return {
            'type': 'fixed',
            'value': normalize_default(path)
        }
    if path['type'] == 'date':
        return {
            'type': 'date',
            'value': normalize_default(path)
        }
    if path['type'] == 'range':
        return {
            'type': 'range',
            'value': normalize_range(path)
            }
    return None


def normalize_url_path(path):
    if path['type'] == 'default':
        return normalize_default(path)
    if path['type'] == 'range':
        return normalize_range(path)
    return '[...]'


def normalize_url_query_param(x, is_first=False):
    prefix = query_params_prefix(is_first).format(x['name'])
    return prefix + normalize_url_path(x)


def normalize_query_param(x, is_first=False):
    first_fragment = {
        'type': 'fixed',
        'value': query_params_prefix(is_first).format(x['name'])
    }
    return [first_fragment, normalize_path(x)]


def normalize_default(x):
    return str(x['values'][0])


def normalize_range(x):
    a, b = x['values'][0:2]
    b_inclusive = str(int(b) - 1)
    return '{}-{}'.format(a, b_inclusive)


def query_params_prefix(is_first):
    return '?{}=' if is_first else '&{}='



"""
Link extraction for auto scraping
"""
from scrapy.link import Link
from scrapy.selector import Selector

from slybot.linkextractor.base import BaseLinkExtractor

class XmlLinkExtractor(BaseLinkExtractor):
    """Link extractor for XML sources"""
    def __init__(self, xpath, **kwargs):
        self.remove_namespaces = kwargs.pop('remove_namespaces', False)
        super(XmlLinkExtractor, self).__init__(**kwargs)
        self.xpath = xpath

    def _extract_links(self, response):
        type = 'html'
        if response.body_as_unicode().strip().startswith('<?xml version='):
            type = 'xml'
        xxs = Selector(response, type=type)
        if self.remove_namespaces:
            xxs.remove_namespaces()
        for url in xxs.xpath(self.xpath).extract():
            yield Link(url.encode(response.encoding))

class RssLinkExtractor(XmlLinkExtractor):
    """Link extraction from RSS feeds"""
    def __init__(self, **kwargs):
        super(RssLinkExtractor, self).__init__("//item/link/text()", **kwargs)

class SitemapLinkExtractor(XmlLinkExtractor):
    """Link extraction for sitemap.xml feeds"""
    def __init__(self, **kwargs):
        kwargs['remove_namespaces'] = True
        super(SitemapLinkExtractor, self).__init__("//urlset/url/loc/text() | //sitemapindex/sitemap/loc/text()", **kwargs)

class AtomLinkExtractor(XmlLinkExtractor):
     def __init__(self, **kwargs):
        kwargs['remove_namespaces'] = True
        super(AtomLinkExtractor, self).__init__("//link/@href", **kwargs)



"""Images."""
from scrapely.extractors import extract_image_url
from slybot.fieldtypes.url import UrlFieldTypeProcessor


class ImagesFieldTypeProcessor(UrlFieldTypeProcessor):
    name = 'image'
    description = 'extracts image URLs'

    def extract(self, text):
        if text is not None:
            return extract_image_url(text) or ''
        return ''



"""
Spider middleware for AS for completing the work made by AS with a "spiderlet" code

"""
from __future__ import absolute_import
import pkgutil, inspect

from scrapy.xlib.pydispatch import dispatcher
from scrapy import signals
from scrapy.exceptions import NotConfigured
from scrapy.http import Request


class DefaultSpiderlet(object):
    name = None

    def __init__(self, spider):
        self.spider = spider

    def process_request(self, request, response):
        return request

    def process_item(self, item, response):
        return item

    def process_start_request(self, request):
        return request

    def parse_login_page(self, response):
        return self.spider.parse_login_page(response)


def list_spiderlets(spiderlets_module_path):
    spiderlets_module = __import__(spiderlets_module_path, {}, {}, [''])
    seen_classes = set()
    for _, mname, _ in pkgutil.iter_modules(spiderlets_module.__path__):
        module = __import__(".".join([spiderlets_module_path, mname]), {}, {}, [''])
        for cls in [c for c in vars(module).values() if inspect.isclass(c)]:
            if cls in seen_classes:
                continue
            seen_classes.add(cls)
            name = getattr(cls, 'name', None)
            if name:
                yield cls


def _load_spiderlet(spiderlets_module_path, spider):
    for cls in list_spiderlets(spiderlets_module_path):
        if cls.name == spider.name:
            class _spiderlet_cls(cls, DefaultSpiderlet):
                pass
            spider.log("SpiderletMiddleware: loaded %s" % _spiderlet_cls.name)
            return _spiderlet_cls(spider)
    return DefaultSpiderlet(spider)


class SpiderletsMiddleware(object):
    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler.settings)

    def __init__(self, settings):
        self.annotating = "annotating" in settings.getlist('SHUB_JOB_TAGS')
        self.spiderlets_module_path = settings["SPIDERLETS_MODULE"]
        if not self.spiderlets_module_path:
            raise NotConfigured
        dispatcher.connect(self.spider_opened, signals.spider_opened)

    def spider_opened(self, spider):
        self.spiderlet = _load_spiderlet(self.spiderlets_module_path, spider)

    def process_spider_output(self, response, result, spider):
        for item_or_request in result:
            if isinstance(item_or_request, Request):
                yield self.spiderlet.process_request(item_or_request, response)
            else:
                yield self.spiderlet.process_item(item_or_request, response)

    def process_start_requests(self, start_requests, spider):
        for request in start_requests:
            if request.callback == spider.parse_login_page:
                request.callback = self.spiderlet.parse_login_page
            yield self.spiderlet.process_start_request(request)



"""
Duplicates filter middleware for autoscraping
"""
from scrapy.exceptions import NotConfigured
from scrapy.exceptions import DropItem

from slybot.item import create_item_version


class DupeFilterPipeline(object):
    def __init__(self, settings):
        if not settings.getbool('SLYDUPEFILTER_ENABLED'):
            raise NotConfigured
        self._itemversion_cache = {}

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler.settings)

    def process_item(self, item, spider):
        """Checks whether a scrapy item is a dupe, based on version (not vary)
        fields of the item class"""
        if (not hasattr(item, 'version_fields') or not item.version_fields or
                item.get('_type') != getattr(item, '_display_name', 0)):
            return item
        version = create_item_version(item)
        if version in self._itemversion_cache:
            old_url = self._itemversion_cache[version]
            raise DropItem("Duplicate product scraped at <%s>, first one was "
                           "scraped at <%s>" % (item["url"], old_url))
        self._itemversion_cache[version] = item["url"]
        return item



import re
from urlparse import urljoin
from scrapely.extractors import url as strip_url
from scrapy.utils.url import safe_download_url
from scrapy.utils.markup import unquote_markup
from slybot.baseurl import get_base_url
disallowed = re.compile('[\x00-\x1F\x7F]')


class UrlFieldTypeProcessor(object):
    """Renders URLs as links"""

    name = 'url'
    description = 'URL'
    limit = 80

    def extract(self, text):
        if text is not None:
            return strip_url(text)
        return ''

    def adapt(self, text, htmlpage=None):
        if htmlpage is None:
            return text
        if text is None:
            return
        encoding = getattr(htmlpage, 'encoding', 'utf-8')
        text = text.encode(encoding)
        unquoted = unquote_markup(text, encoding=encoding)
        cleaned = strip_url(disallowed.sub('', unquoted))
        base = get_base_url(htmlpage).encode(encoding)
        base_url = strip_url(unquote_markup(base, encoding=encoding))
        joined = urljoin(base_url, cleaned)
        return safe_download_url(joined)



from __future__ import absolute_import

import itertools
import json

from copy import deepcopy

from loginform import fill_login_form

from scrapy.http import FormRequest, HtmlResponse, Request, XmlResponse
from scrapy.spiders.sitemap import SitemapSpider
from scrapy.utils.spider import arg_to_iter

import six
from six.moves.urllib_parse import urlparse

from slybot.generic_form import GenericForm
from slybot.linkextractor import create_linkextractor_from_specs
from slybot.starturls import (
    FragmentGenerator, FeedGenerator, IdentityGenerator, StartUrlCollection,
    UrlGenerator
)
from slybot.utils import (
    include_exclude_filter, IndexedDict, iter_unique_scheme_hostname,
    load_plugin_names, load_plugins,
)
from w3lib.http import basic_auth_header

STRING_KEYS = ['start_urls', 'exclude_patterns', 'follow_patterns',
               'allowed_domains', 'js_enabled', 'js_enable_patterns',
               'js_disable_patterns']


class IblSpider(SitemapSpider):
    def __init__(self, name, spec, item_schemas, all_extractors, settings=None,
                 **kw):
        self.start_url_generators = {
            'start_urls': IdentityGenerator(),
            'generated_urls': UrlGenerator(settings, kw),

            'url': IdentityGenerator(),
            'feed': FeedGenerator(self.parse),
            'generated': FragmentGenerator(),
        }
        self.generic_form = GenericForm(**kw)
        super(IblSpider, self).__init__(name, **kw)
        spec = deepcopy(spec)
        self._add_spider_args_to_spec(spec, kw)
        self.plugins = self._configure_plugins(
            settings, spec, item_schemas, all_extractors)
        self._configure_js(spec, settings)

        self.login_requests, self.form_requests = [], []
        self._start_urls = self._create_start_urls(spec)
        self._start_requests = self._create_start_requests(spec)
        self._create_init_requests(spec)
        self._add_allowed_domains(spec)
        self.page_actions = spec.get('page_actions', [])

    def _add_spider_args_to_spec(self, spec, args):
        for key, val in args.items():
            if isinstance(val, six.string_types) and key in STRING_KEYS:
                val = val.splitlines()
            spec[key] = val

    def _create_start_urls(self, spec):
        url_type = spec.get('start_urls_type', 'start_urls')
        return StartUrlCollection(
            arg_to_iter(spec[url_type]),
            self.start_url_generators,
        )

    def _create_start_requests(self, spec):
        init_requests = spec.get('init_requests', [])
        for rdata in init_requests:
            if rdata["type"] == "start":
                yield self._create_start_request_from_specs(rdata)

        for start_url in self._start_urls:
            if not isinstance(start_url, Request):
                start_url = Request(start_url, callback=self.parse,
                                    dont_filter=True)
            yield self._add_splash_meta(start_url)

    def _create_init_requests(self, spec):
        init_requests = spec.get('init_requests', [])
        for rdata in init_requests:
            if rdata["type"] == "login":
                request = Request(url=rdata.pop("loginurl"), meta=rdata,
                                  callback=self.parse_login_page,
                                  dont_filter=True)
                self._add_splash_meta(request)
                self.login_requests.append(request)
            elif rdata["type"] == "form":
                self.form_requests.append(
                    self.get_generic_form_start_request(rdata)
                )

    def _add_allowed_domains(self, spec):
        self.allowed_domains = spec.get('allowed_domains', [])
        if self.allowed_domains is not None and not self.allowed_domains:
            self.allowed_domains = self._get_allowed_domains(spec)

    def parse_login_page(self, response):
        username = response.request.meta["username"]
        password = response.request.meta["password"]
        args, url, method = fill_login_form(response.url, response.body,
                                            username, password)
        return FormRequest(url, method=method, formdata=args,
                           callback=self.after_login, dont_filter=True)

    def after_login(self, response):
        for result in self.parse(response):
            yield result
        for req in self._start_requests:
            yield req

    def get_generic_form_start_request(self, form_descriptor):
        file_fields = list(self.generic_form.get_url_field(form_descriptor))
        if file_fields:
            (field_index, field_descriptor) = file_fields.pop(0)
            form_descriptor['field_index'] = field_index
            return FormRequest(self.generic_form.get_value(field_descriptor),
                               meta=form_descriptor,
                               callback=self.parse_field_url_page,
                               dont_filter=True)
        else:
            return Request(url=form_descriptor.pop("form_url"),
                           meta=form_descriptor, callback=self.parse_form_page,
                           dont_filter=True)

    def parse_field_url_page(self, response):
        form_descriptor = response.request.meta
        field_index = form_descriptor['field_index']
        field_descriptor = form_descriptor['fields'][field_index]
        self.generic_form.set_values_url_field(field_descriptor, response.body)
        yield self.get_generic_form_start_request(form_descriptor)

    def parse_form_page(self, response):
        fill_form = self.generic_form.fill_generic_form
        try:
            for (args, url, method) in fill_form(response.url, response.body,
                                                 response.request.meta):
                yield FormRequest(url, method=method, formdata=args,
                                  callback=self.after_form_page,
                                  dont_filter=True)
        except Exception as e:
            self.logger.warning(str(e))
        for req in self._start_requests:
            yield req

    def after_form_page(self, response):
        for result in self.parse(response):
            yield result

    def _get_allowed_domains(self, spec):
        urls = [x['url'] for x in spec['templates']]
        urls += [x['url'] for x in spec.get('init_requests', [])
                 if x['type'] == 'start']
        urls += self._start_urls.allowed_domains
        return [domain for scheme, domain in iter_unique_scheme_hostname(urls)]

    def start_requests(self):
        start_requests = []
        if self.login_requests:
            start_requests = self.login_requests
        elif self.form_requests:
            start_requests = self.form_requests
        else:
            start_requests = self._start_requests
        for req in start_requests:
            yield req

    def _create_start_request_from_specs(self, info):
        url = info["url"]
        lspecs = info.get("link_extractor")
        if lspecs:
            linkextractor = create_linkextractor_from_specs(lspecs)

            def _callback(spider, response):
                for link in linkextractor.links_to_follow(response):
                    request = Request(url=link.url, callback=spider.parse)
                    yield self._add_splash_meta(request)
            request = Request(url=url, callback=_callback)
            return self._add_splash_meta(request)
        request = Request(url=url, callback=self.parse)
        return self._add_splash_meta(request)

    def parse(self, response):
        """Main handler for all downloaded responses"""
        request = response.request
        if (request and request.method == 'POST' and
                urlparse(request.url).hostname == self.SPLASH_HOST):
            url = (json.loads(request.body).get('url'))
            if url:
                response._url = url
        content_type = response.headers.get('Content-Type', '')
        if isinstance(response, HtmlResponse):
            return self.handle_html(response)
        if (isinstance(response, XmlResponse) or
                response.url.endswith(('.xml', '.xml.gz'))):
            response._set_body(self._get_sitemap_body(response))
            return self.handle_xml(response)
        self.logger.debug(
            "Ignoring page with content-type=%r: %s" % (content_type,
                                                        response.url)
        )
        return []

    def _configure_plugins(self, settings, spec, schemas, extractors):
        plugins = IndexedDict()
        for plugin_class, plugin_name in zip(load_plugins(settings),
                                             load_plugin_names(settings)):
            instance = plugin_class()
            instance.setup_bot(settings, spec, schemas, extractors, self.logger)
            plugins[plugin_name] = instance
        return plugins

    def _plugin_hook(self, name, *args):
        results = []
        for plugin in self.plugins.values():
            if hasattr(plugin, name):
                results.append(getattr(plugin, name)(*args))
        return results

    def _handle(self, hook, response, *extrasrgs):
        generators = self._plugin_hook(hook, response, *extrasrgs)
        for item_or_request in itertools.chain(*generators):
            if isinstance(item_or_request, Request):
                self._plugin_hook('process_request', item_or_request, response)
            else:
                self._plugin_hook('process_item', item_or_request, response)
            if isinstance(item_or_request, Request):
                item_or_request = self._add_splash_meta(item_or_request)
            yield item_or_request

    def handle_xml(self, response):
        return self._handle('handle_xml', response, set([]))

    def handle_html(self, response):
        return self._handle('handle_html', response)

    def _configure_js(self, spec, settings):
        self.js_enabled = False
        self.SPLASH_HOST = None
        if settings.get('SPLASH_URL'):
            self.SPLASH_HOST = urlparse(settings.get('SPLASH_URL')).hostname
            self.js_enabled = spec.get('js_enabled', False)
        if self.js_enabled and (settings.get('SPLASH_PASS') is not None or
                                settings.get('SPLASH_USER') is not None):
            self.splash_auth = basic_auth_header(
                settings.get('SPLASH_USER', ''),
                settings.get('SPLASH_PASS', ''))
        self.splash_wait = settings.getint('SPLASH_WAIT', 5)
        self.splash_timeout = settings.getint('SPLASH_TIMEOUT', 30)
        self.splash_js_source = settings.get(
            'SPLASH_JS_SOURCE', 'function(){}')
        self.splash_lua_source = settings.get('SPLASH_LUA_SOURCE', '')
        self._filter_js_urls = self._build_js_url_filter(spec)

    def _build_js_url_filter(self, spec):
        if not self.js_enabled:
            return lambda x: None
        enable_patterns = spec.get('js_enable_patterns')
        disable_patterns = spec.get('js_disable_patterns')
        return include_exclude_filter(enable_patterns, disable_patterns)

    def _add_splash_meta(self, request):
        if self.js_enabled and self._filter_js_urls(request.url):
            cleaned_url = urlparse(request.url)._replace(params='', query='',
                                                         fragment='').geturl()
            endpoint = 'execute' if self.splash_lua_source else 'render.html'
            request.meta['splash'] = {
                'endpoint': endpoint,
                'args': {
                    'wait': self.splash_wait,
                    'timeout': self.splash_timeout,
                    'js_source': self.splash_js_source,
                    'lua_source': self.splash_lua_source,
                    'images': 0,
                    'url': request.url,
                    'baseurl': cleaned_url
                }
            }
        return request



from __future__ import absolute_import

import itertools
import operator
import re

from collections import OrderedDict

from scrapy.http import Request
from scrapy.selector import Selector
from scrapy.utils.misc import arg_to_iter

from scrapely.extraction import InstanceBasedLearningExtractor
from scrapely.htmlpage import HtmlPage, dict_to_page

from slybot.linkextractor import create_linkextractor_from_specs
from slybot.linkextractor.html import HtmlLinkExtractor
from slybot.linkextractor.xml import SitemapLinkExtractor
from slybot.linkextractor.pagination import PaginationExtractor
from slybot.item import SlybotItem, create_slybot_item_descriptor
from slybot.extractors import apply_extractors, add_extractors_to_descriptors
from slybot.utils import (htmlpage_from_response, include_exclude_filter,
                          _build_sample)
from .extraction import SlybotIBLExtractor
XML_APPLICATION_TYPE = re.compile('application/((?P<type>[a-z]+)\+)?xml').match
_CLUSTER_NA = 'not available'
_CLUSTER_OUTLIER = 'outlier'


class Annotations(object):
    """
    Base Class for adding plugins to Portia Web and Slybot.
    """

    def setup_bot(self, settings, spec, items, extractors, logger):
        """
        Perform any initialization needed for crawling using this plugin
        """
        self.logger = logger
        templates = map(self._get_annotated_template, spec['templates'])

        _item_template_pages = sorted((
            [t.get('scrapes'), dict_to_page(t, 'annotated_body'),
             t.get('extractors', []), t.get('version', '0.12.0')]
            for t in templates if t.get('page_type', 'item') == 'item'
        ), key=lambda x: x[0])
        self.item_classes = {}
        self.template_scrapes = {template.get('page_id'): template['scrapes']
                                 for template in templates}
        if (settings.get('AUTO_PAGINATION') or
                spec.get('links_to_follow') == 'auto'):
            self.html_link_extractor = PaginationExtractor()
        else:
            self.html_link_extractor = HtmlLinkExtractor()
        for schema_name, schema in items.items():
            if schema_name not in self.item_classes:
                if not schema.get('name'):
                    schema['name'] = schema_name
                item_cls = SlybotItem.create_iblitem_class(schema)
                self.item_classes[schema_name] = item_cls

        # Create descriptors and apply additional extractors to fields
        page_descriptor_pairs = []
        self.schema_descriptors = {}
        for default, template, template_extractors, v in _item_template_pages:
            descriptors = OrderedDict()
            for schema_name, schema in items.items():
                item_descriptor = create_slybot_item_descriptor(schema,
                                                                schema_name)
                apply_extractors(item_descriptor, template_extractors,
                                 extractors)
                descriptors[schema_name] = item_descriptor
            descriptor = descriptors.values() or [{}]
            descriptors['#default'] = descriptors.get(default, descriptor[0])
            self.schema_descriptors[template.page_id] = descriptors['#default']
            page_descriptor_pairs.append((template, descriptors, v))
            add_extractors_to_descriptors(descriptors, extractors)

        grouped = itertools.groupby(sorted(page_descriptor_pairs,
                                           key=operator.itemgetter(2)),
                                    lambda x: x[2] < '0.13.0')
        self.extractors = []
        for version, group in grouped:
            if version:
                self.extractors.append(
                    InstanceBasedLearningExtractor(
                        [(page, scrapes['#default'])
                         for page, scrapes, version in group]))
            else:
                self.extractors.append(SlybotIBLExtractor(list(group)))

        # generate ibl extractor for links pages
        _links_pages = [dict_to_page(t, 'annotated_body')
                        for t in templates if t.get('page_type') == 'links']
        _links_item_descriptor = create_slybot_item_descriptor({'fields': {}})
        self._links_ibl_extractor = InstanceBasedLearningExtractor(
            [(t, _links_item_descriptor) for t in _links_pages]) \
            if _links_pages else None

        self.build_url_filter(spec)
        # Clustering
        self.template_names = [t.get('page_id') for t in spec['templates']]
        if settings.get('PAGE_CLUSTERING'):
            try:
                import page_clustering
                self.clustering = page_clustering.kmeans_from_samples(spec['templates'])
                self.logger.info("Clustering activated")
            except ImportError:
                self.clustering = None
                self.logger.warning(
                    "Clustering could not be used because it is not installed")
        else:
            self.clustering = None

    def _get_annotated_template(self, template):
        if template.get('version', '0.12.0') >= '0.13.0':
            _build_sample(template)
        return template

    def handle_html(self, response, seen=None):
        htmlpage = htmlpage_from_response(response, _add_tagids=True)
        items, link_regions = self.extract_items(htmlpage, response)
        htmlpage.headers['n_items'] = len(items)
        try:
            response.meta['n_items'] = len(items)
        except AttributeError:
            pass  # response not tied to any request
        for item in items:
            yield item
        for request in self._process_link_regions(htmlpage, link_regions):
            yield request

    def extract_items(self, htmlpage, response=None):
        """This method is also called from UI webservice to extract items"""
        for extractor in self.extractors:
            items, links = self._do_extract_items_from(htmlpage, extractor,
                                                       response)
            if items:
                return items, links
        return [], []

    def _do_extract_items_from(self, htmlpage, extractor, response=None):
        # Try to predict template to use
        template_cluster, pref_template_id = self._cluster_page(htmlpage)
        extracted_data, template = extractor.extract(htmlpage, pref_template_id)
        extracted = []
        sel = Selector(text=htmlpage.body)
        for item in arg_to_iter(extracted_data):
            try:
                extracted.append(item.process(sel))
            except AttributeError:
                extracted.append(item)
        link_regions = []
        for ddict in extracted or []:
            link_regions.extend(arg_to_iter(ddict.pop("_links", [])))
        descriptor = None
        unprocessed = False
        if template is not None and hasattr(template, 'descriptor'):
            descriptor = template.descriptor()
            if hasattr(descriptor, 'name'):
                item_cls_name = descriptor.name
            elif hasattr(descriptor, 'get'):
                item_cls_name = descriptor.get('name',
                                               descriptor.get('display_name'))
            else:
                item_cls_name = ''
        else:
            unprocessed = True
            try:
                descriptor = self.schema_descriptors[template.id]
                item_cls_name = self.template_scrapes[template.id]
            except AttributeError:
                descriptor = sorted(self.schema_descriptors.items())[0][1]
                item_cls_name = sorted(self.template_scrapes.items())[0][1]
        item_cls = self.item_classes.get(item_cls_name)
        items = []
        for processed_attributes in extracted:
            if processed_attributes.get('_type') in self.item_classes:
                _type = processed_attributes['_type']
                item = self.item_classes[_type](processed_attributes)
                item['_type'] = item.display_name()
            elif unprocessed:
                item = self._process_attributes(processed_attributes,
                                                descriptor, htmlpage)
                if item_cls:
                    item = item_cls(item)
            elif item_cls:
                item = item_cls(processed_attributes)
            else:
                item = dict(processed_attributes)
            item[u'url'] = htmlpage.url
            item[u'_template'] = str(template.id)
            item.setdefault('_type', item_cls_name)
            if not isinstance(item, SlybotItem):
                default_meta = {'type': 'text', 'required': False,
                                'vary': False}
                item_cls = SlybotItem.create_iblitem_class(
                    {'fields': {k: default_meta for k in item}}
                )
                item = item_cls(**item)
            if self.clustering:
                item['_template_cluster'] = template_cluster
            items.append(item)
        return items, link_regions

    def _process_attributes(self, item, descriptor, htmlpage):
        new_item = {}
        try:
            attr_map = descriptor.attribute_map
        except AttributeError:
            attr_map = {}
        page = getattr(htmlpage, 'htmlpage', htmlpage)
        for field, value in item.items():
            if field.startswith('_sticky'):
                continue
            if field == 'variants':
                value = [self._process_attributes(v, descriptor, page)
                         for v in value]
            elif field in attr_map:
                value = [attr_map[field].adapt(v, page) for v in value]
            new_item[field] = value
        return new_item

    def build_url_filter(self, spec):
        """make a filter for links"""
        respect_nofollow = spec.get('respect_nofollow', True)

        if spec.get("links_to_follow") == "none":
            url_filterf = lambda x: False
        elif spec.get("links_to_follow") == "all":
            if respect_nofollow:
                url_filterf = lambda x: x.nofollow
            else:
                url_filterf = lambda x: True
        else: # patterns
            patterns = spec.get('follow_patterns')
            excludes = spec.get('exclude_patterns')
            pattern_fn = include_exclude_filter(patterns, excludes)

            if respect_nofollow:
                url_filterf = lambda x: not x.nofollow and pattern_fn(x.url)
            else:
                url_filterf = lambda x: pattern_fn(x.url)

        self.url_filterf = url_filterf

    def _cluster_page(self, htmlpage):
        template_cluster, preferred = _CLUSTER_NA, None
        if self.clustering:
            self.clustering.add_page(htmlpage)
            if self.clustering.is_fit:
                clt = self.clustering.classify(htmlpage)
                if clt != -1:
                    template_cluster = preferred = self.template_names[clt]
                else:
                    template_cluster = _CLUSTER_OUTLIER
        return template_cluster, preferred


    def _filter_link(self, link, seen):
        url = link.url
        if self.url_filterf(link):
            # filter out duplicate urls, later we should handle link text
            if url not in seen:
                seen.add(url)
                request = Request(url)
                if link.text:
                    request.meta['link_text'] = link.text
                return request

    def _process_link_regions(self, htmlpage, link_regions):
        """Process link regions if any, and generate requests"""
        if link_regions:
            for link_region in link_regions:
                htmlregion = HtmlPage(htmlpage.url, htmlpage.headers,
                                      link_region, encoding=htmlpage.encoding)
                for request in self._requests_to_follow(htmlregion):
                    yield request
        else:
            for request in self._requests_to_follow(htmlpage):
                yield request

    def _requests_to_follow(self, htmlpage):
        if self._links_ibl_extractor is not None:
            extracted = self._links_ibl_extractor.extract(htmlpage)[0]
            if extracted:
                extracted_regions = extracted[0].get('_links', [])
                seen = set()
                for region in extracted_regions:
                    htmlregion = HtmlPage(htmlpage.url, htmlpage.headers,
                                          region, encoding=htmlpage.encoding)
                    for request in self._request_to_follow_from_region(
                            htmlregion):
                        if request.url in seen:
                            continue
                        seen.add(request.url)
                        yield request
        else:
            for request in self._request_to_follow_from_region(htmlpage):
                yield request

    def _request_to_follow_from_region(self, htmlregion):
        seen = set()
        for link in self.html_link_extractor.links_to_follow(htmlregion):
            request = self._filter_link(link, seen)
            if request is not None:
                yield request

    def handle_xml(self, response, seen):
        _type = XML_APPLICATION_TYPE(response.headers.get('Content-Type', ''))
        _type = _type.groupdict()['type'] if _type else 'xml'
        try:
            link_extractor = create_linkextractor_from_specs({
                'type': _type, 'value': ''
            })
        except ValueError:
            link_extractor = SitemapLinkExtractor()
        for link in link_extractor.links_to_follow(response):
            request = self._filter_link(link, seen)
            if request:
                yield request



import json

from scrapy import Selector
from scrapely.htmlpage import parse_html, HtmlTag, HtmlDataFragment

from collections import defaultdict
from itertools import tee, count, groupby
from operator import itemgetter

from slybot.utils import (serialize_tag, add_tagids, remove_tagids, TAGID,
                          OPEN_TAG, CLOSE_TAG, UNPAIRED_TAG, GENERATEDTAGID)

from .migration import _get_parent, short_guid


class Annotations(object):

    def save_extraction_data(self, data, template, options={}):
        """
        data = {
            extracts: [
                {
                    annotations: {"content": "Title"},
                    id: "id-string",
                    required: [],
                    tagid: 12,
                    # All keys below are optional
                    variant: 0,
                    text-content: "name-of-text-content-field",
                    ignore: True,
                    ignore_beneath: True,
                    insert_after: True,
                    slice: [2, 16],
                    item_container: True,
                    container_id: "parent-id-string",
                    schema_id: "schema-id-string",
                    repeated: true,
                    siblings: 2,
                    field: "field-id-to-be-added-to-in-parent-container"
                }
            ]
        }
        """
        annotation_data = _clean_annotation_data(data.get('extracts', []))
        data['extracts'] = annotation_data
        template['annotated_body'] = apply_annotations(
            annotation_data,
            template[options.get('body', 'original_body')])
        return data


def _clean_annotation_data(data):
    result = []
    sticky_count, stickies = count(1), set()
    for ann in data:
        if ann.get('item_container'):
            ann['annotations'] = {'#portia-content': '#dummy'}
            ann['text-content'] = '#portia-content'
        elif 'data' in ann:
            modified_annotations = {}
            grp = itemgetter('attribute')
            for _id, value in ann['data'].items():
                value['id'] = '%s|%s' % (ann['id'], _id)
            sorted_annotations = sorted(ann['data'].values(), key=grp)
            for attribute, annotations in groupby(sorted_annotations, grp):
                modified_annotations[attribute] = list(annotations)
            ann['annotations'] = modified_annotations
        elif 'annotations' in ann:
            filtered_annotations = {}
            for k, v in ann['annotations'].items():
                if not v:
                    continue
                if v == '#sticky':
                    next_sticky = '_sticky%s' % next(sticky_count)
                    stickies.add(next_sticky)
                    v = next_sticky
                filtered_annotations[k] = v

            ann['annotations'] = filtered_annotations
            ann['required'] = list((set(ann.get('required', [])) | stickies) &
                                   set(filtered_annotations.values()))
        elif "ignore" in ann or "ignore_beneath" in ann:
            pass
        else:
            continue
        result.append(ann)
    return result


def _get_data_id(annotation):
    """Get id (a str) of an annotation."""
    if isinstance(annotation, HtmlTag):
        return annotation.attributes[TAGID]


def _gen_annotation_info(annotation):
    data = {}
    if 'annotations' in annotation:
        data['data-scrapy-annotate'] = json.dumps({
            'id': annotation.get('id', short_guid()),
            'annotations': annotation.get('annotations', {}),
            'required': annotation.get('required', []),
            'required_fields': annotation.get('required', []),
            'variant': int(annotation.get('variant', 0)),
            'generated': annotation.get('generated', False),
            'text-content': annotation.get('text-content', 'content'),
            'item_container': annotation.get('item_container', False),
            'container_id': annotation.get('container_id'),
            'schema_id': annotation.get('schema_id'),
            'repeated': annotation.get('repeated'),
            'siblings': annotation.get('siblings'),
            'field': annotation.get('field'),
            'selector': annotation.get('selector'),
            'selection_mode': annotation.get('selection_mode'),
            'min_jump': annotation.get('min_jump', -1),
            'max_separator': annotation.get('max_separator', -1)
        }).replace('"', '&quot;')
    if 'ignore' in annotation or 'ignore_beneath' in annotation:
        if annotation.get('ignore_beneath'):
            data['data-scrapy-ignore-beneath'] = 'true'
        elif annotation.get('ignore'):
            data['data-scrapy-ignore'] = 'true'
    return data


def _get_generated_annotation(element, annotations, nodes, html_body, inserts):
    eid = insert_after_tag = _get_data_id(element)
    text_strings = _get_text_nodes(nodes, html_body)
    text_content = ''.join((s.lstrip() for s in text_strings))
    pre_selected = []
    for annotation in annotations:
        start, end = _get_generated_slice(annotation)
        pre_selected.append((text_content[0:start], text_content[start:end],
                             annotation))
    tag_stack = [insert_after_tag]
    next_text_node = ''
    for i, node in enumerate(nodes):
        if isinstance(node, HtmlTag):
            if node.tag_type == OPEN_TAG:
                tagid = node.attributes.get(TAGID, '').strip()
                if tagid:
                    tag_stack.append(tagid)
            elif node.tag_type == CLOSE_TAG and tag_stack:
                insert_after_tag = tag_stack.pop()
        elif (isinstance(node, HtmlDataFragment) and len(tag_stack) == 1):
            text = html_body[node.start:node.end]
            # This allows for a clean way to insert fragments up until the
            # next tag in apply_annotations if we have already inserted a new
            # generated tag
            if not node.is_text_content and inserts.get(insert_after_tag):
                inserts[insert_after_tag].append(text)
                continue
            removed = 0
            inserted = False
            for j, (pre, selected, annotation) in enumerate(pre_selected[:]):
                if selected and selected in text:
                    previous, post = text.split(selected, 1)
                    if previous.strip() in pre:
                        pre_selected.pop(j - removed)
                        removed += 1
                        generated = _generate_elem(annotation, selected)
                        # Next immediate text node will be returned and added
                        # to the new document. Other text nodes within this
                        # node will be added after other child nodes have been
                        # closed.
                        if (insert_after_tag == eid and
                                not annotation.get('insert_after')):
                            next_text_node += previous + generated
                            inserted = True
                        else:
                            inserts[insert_after_tag].extend([previous,
                                                              generated])
                        text = post
            if inserted:
                next_text_node += text
            else:
                inserts[insert_after_tag].append(text)
    return next_text_node


def _get_text_nodes(nodes, html_body):
    text = []
    open_tags = 0
    for node in nodes:
        if isinstance(node, HtmlTag):
            if node.tag_type == OPEN_TAG:
                open_tags += 1
            elif node.tag_type == CLOSE_TAG:
                open_tags -= 1
        elif (isinstance(node, HtmlDataFragment) and
              node.is_text_content and open_tags == 0):
            text.append(html_body[node.start:node.end])
    return text


def _get_generated_slice(annotation):
    annotation_slice = annotation.get('slice', [0])[:2]
    if not annotation_slice:
        annotation_slice = [0, 0]
    elif len(annotation_slice) < 2:
        annotation_slice.append(annotation_slice[0])
    return annotation_slice


def _generate_elem(annotation, text):
    sections = ['<ins']
    annotation_info = _gen_annotation_info(annotation)
    annotation_info[GENERATEDTAGID] = annotation.get('id')
    attributes = []
    for key, value in annotation_info.items():
        attributes.append('%s="%s"' % (key, value))
    sections.append(' '.join(attributes))
    if len(sections) > 1:
        sections[0] += ' '
    sections.extend(['>', text, '</ins>'])
    return ''.join(sections)


def _get_inner_nodes(target, open_tags=1, insert_after=False,
                     stop_on_next=False):
    nodes = []
    while open_tags > -0:
        elem = next(target)
        if isinstance(elem, HtmlTag):
            if elem.tag_type == OPEN_TAG:
                open_tags += 1
                if stop_on_next and elem.attributes.get(TAGID) is not None:
                    return nodes
            elif (stop_on_next and
                  elem.tag_type == UNPAIRED_TAG and
                  elem.attributes.get(TAGID) is not None):
                return nodes
            elif elem.tag_type == CLOSE_TAG:
                open_tags -= 1
        nodes.append(elem)
    if insert_after:
        return _get_inner_nodes(target, stop_on_next=True)
    return nodes


def _add_element(element, output, html):
    if '__added' not in element.attributes:
        output.append(html[element.start:element.end])
        element.attributes['__added'] = True
    return element


def _annotation_key(a):
    return a.get('generated', False) + sum(a.get('slice', []))


def _filter_annotations(annotations):
    selector, tagid = [], []
    for ann in annotations:
        if ann:
            if ann.get('selector'):
                selector.append(ann)
            elif ann.get('tagid') and (ann.get('annotations') or
                                       ann.get('ignore')):
                tagid.append(ann)
    return selector, tagid


def _merge_annotations_by_selector(annotations):
    def grouper(x):
        return x.get('selector')
    annotations.sort(key=grouper)
    new_annotations = []
    for sel, annos in groupby(annotations, key=grouper):
        annos = list(annos)
        anno = annos[0]
        if len(annos) == 1:
            new_annotations.append(anno)
            continue
        for other_anno in annos[1:]:
            # TODO: Handle annotations pointing to different containers
            for attribute, data in other_anno['annotations'].items():
                if attribute in anno['annotations']:
                    attr_data = anno['annotations'][attribute]
                    if isinstance(attr_data, list):
                        attr_data.extend(data)
                    else:
                        current = {'field': attr_data, 'attribute': attribute,
                                   'required': False, 'extractors': []}
                        attr_data = [current, data]
                else:
                    anno['annotations'][attribute] = data
        new_annotations.append(anno)
    return new_annotations


def apply_selector_annotations(annotations, target_page):
    page = Selector(text=target_page)
    converted_annotations = []
    annotations = _merge_annotations_by_selector(annotations)
    tagid_selector_map = {}
    added_repeated = {}
    containers = {}
    for annotation in annotations:
        if annotation.get('item_container'):
            containers[annotation['id']] = annotation
        selector = annotation.get('selector')
        tagid, elems = tagid_for_annotation(annotation, page)
        if tagid is not None:
            annotation['tagid'] = tagid
            if selector:
                tagid_selector_map[tagid] = selector
            converted_annotations.append(annotation)

        # Create container for repeated field annotation
        if (annotation.get('repeated') and
                not annotation.get('item_container') and
                len(annotation.get('annotations')) == 1):
            repeated_parent = add_repeated_field(annotation, elems, page)
            if repeated_parent:
                converted_annotations.append(repeated_parent)
                container_id = repeated_parent['container_id']
                added_repeated[container_id] = repeated_parent
    if added_repeated:
        for container_id, child in added_repeated.items():
            container = containers[container_id]
            if container['tagid'] != child['tagid']:
                continue
            _, elems = tagid_for_annotation(container, page)
            parent = elems[0].getparent()
            container['tagid'] = int(parent.attrib.get('data-tagid', 1e9))
    return converted_annotations


def tagid_for_annotation(annotation, page):
    selector = annotation.get('selector')
    if not selector:
        return None, None
    elems = []
    while not elems:
        elems = [elem._root for elem in page.css(selector)]
        selector = ' > '.join(selector.split(' > ')[1:])
    if not elems:
        return None, None

    tagids = [int(e.attrib.get('data-tagid', 1e9)) for e in elems]
    return min(tagids), elems


def add_repeated_field(annotation, elems, page):
    parent = _get_parent(elems, page)
    field = annotation['annotations'].values()[0][0]['field']
    container_id = '%s#parent' % annotation['id']
    if len(parent):
        tagid = int(parent.attrib.get('data-tagid', 1e9))
        parent_annotation = {
            'item_container': True,
            'id': container_id,
            'annotations': {'#portia-content': '#dummy'},
            'text-content': '#portia-content',
            'container_id': annotation['container_id'],
            'field': field,
            'tagid': tagid
        }
        annotation['item_container'] = True
        annotation['field'] = field
        annotation['container_id'] = container_id
        return parent_annotation


def apply_annotations(annotations, target_page):
    selector_annotations, tagid_annotations = _filter_annotations(annotations)
    inserts = defaultdict(list)
    numbered_html = add_tagids(target_page)
    if selector_annotations:
        converted_annotations = apply_selector_annotations(
            selector_annotations, numbered_html)
        tagid_annotations += converted_annotations
    target = iter(parse_html(numbered_html))
    output, tag_stack = [], []
    element = next(target)
    last_id = 0
    # XXX: A dummy element is added to the end so if the last annotation is
    #      generated it will be added to the output
    filtered = defaultdict(list)
    for ann in tagid_annotations:
        filtered[ann['tagid']].append(ann)
    dummy = [(1e9, [{}])]
    sorted_annotations = sorted([(int(k), v) for k, v in filtered.items()] +
                                dummy)
    try:
        for aid, annotation_data in sorted_annotations:
            # Move target until replacement/insertion point
            while True:
                while not isinstance(element, HtmlTag) or element.tag == 'ins':
                    output.append(numbered_html[element.start:element.end])
                    element = next(target)
                if element.tag_type in {OPEN_TAG, UNPAIRED_TAG}:
                    last_id = element.attributes.get(TAGID)
                    tag_stack.append(last_id)
                if element.tag_type in {CLOSE_TAG, UNPAIRED_TAG} and tag_stack:
                    if ('__added' not in element.attributes and
                            last_id is not None and aid is not None and
                            int(last_id) < int(aid)):
                        output.append(numbered_html[element.start:element.end])
                        element.attributes['__added'] = True
                    last_inserted = tag_stack.pop()
                    to_insert = inserts.pop(last_inserted, None)
                    if to_insert:
                        output.extend(to_insert)
                        # Skip all nodes up to the next HtmlTag as these
                        # have already been added
                        while True:
                            element = next(target)
                            try:
                                last_id = element.attributes.get(TAGID,
                                                                 last_id)
                            except AttributeError:
                                pass
                            if isinstance(element, HtmlTag):
                                break
                        continue
                if (last_id is not None and aid is not None and
                        int(last_id) < int(aid)):
                    if '__added' not in element.attributes:
                        output.append(numbered_html[element.start:element.end])
                        element.attributes['__added'] = True
                    element = next(target)
                else:
                    break

            generated = []
            next_generated = []
            # Place generated annotations at the end and sort by slice
            for annotation in sorted(annotation_data, key=_annotation_key):
                if annotation.get('generated'):
                    if annotation.get('insert_after'):
                        next_generated.append(annotation)
                    else:
                        generated.append(annotation)
                else:
                    # Add annotations data as required
                    annotation_info = _gen_annotation_info(annotation)
                    for key, val in annotation_info.items():
                        element.attributes[key] = val
            next_text_section = ''
            if generated:
                inner_data, target = tee(target)
                nodes = _get_inner_nodes(inner_data)
                next_text_section = _get_generated_annotation(
                    element, generated, nodes, numbered_html, inserts)
            if next_generated:
                inner_data, target = tee(target)
                open_tags = 0 if element.tag_type == UNPAIRED_TAG else 1
                nodes = _get_inner_nodes(inner_data, open_tags=open_tags,
                                         insert_after=True)
                next_text_section = _get_generated_annotation(
                    element, next_generated, nodes, numbered_html, inserts)

            if '__added' not in element.attributes:
                output.append(serialize_tag(element))
                element.attributes['__added'] = True
            # If an <ins> tag has been inserted we need to move forward
            if next_text_section:
                while True:
                    elem = next(target)
                    if (isinstance(elem, HtmlDataFragment) and
                            elem.is_text_content):
                        break
                    output.append(numbered_html[elem.start:elem.end])
                output.append(next_text_section)
    # Reached the end of the document
    except StopIteration:
        output.append(numbered_html[element.start:element.end])
    else:
        for element in target:
            output.append(numbered_html[element.start:element.end])
    return remove_tagids(''.join(output))



from unittest import TestCase
from os.path import dirname

from scrapy.http import HtmlResponse
from scrapy.settings import Settings
from scrapy.item import DictItem
from scrapy.exceptions import DropItem

from slybot.spidermanager import SlybotSpiderManager
from slybot.dupefilter import DupeFilterPipeline

_PATH = dirname(__file__)

class DupeFilterTest(TestCase):
    smanager = SlybotSpiderManager("%s/data/SampleProject" % _PATH)

    def test_dupefilter(self):
        name = "seedsofchange2"
        spider = self.smanager.create(name)
        spec = self.smanager._specs["spiders"][name]
        t1, t2 = spec["templates"]

        dupefilter = DupeFilterPipeline(Settings({"SLYDUPEFILTER_ENABLED": True}))

        response1 = HtmlResponse(url=t1["url"], body=t1["original_body"].encode('utf-8'))
        response2 = HtmlResponse(url=t2["url"], body=t2["original_body"].encode('utf-8'))

        result1 = spider.handle_html(response1)
        for item1 in result1:
            if isinstance(item1, DictItem):
                break

        result2 = spider.handle_html(response2)
        for item2 in result2:
            if isinstance(item2, DictItem):
                break

        self.assertEqual(item1, dupefilter.process_item(item1, spider))
        self.assertEqual(item2, dupefilter.process_item(item2, spider))

        self.assertRaises(DropItem, dupefilter.process_item, item1, spider)



"""Simple validation of specifications passed to slybot"""
from __future__ import absolute_import
from os.path import dirname, join
import json
import re
import socket

from urlparse import urlparse, parse_qsl
from urllib import urlencode
from urllib2 import unquote
from six.moves.urllib.parse import urlsplit, urlunsplit

from jsonschema import (Draft3Validator, RefResolver, FormatChecker,
                        ValidationError)
import six


_PATH = dirname(__file__)


def load_schemas():
    filename = join(_PATH, "schemas.json")
    return dict((s["id"], s) for s in json.load(open(filename)))

_SCHEMAS = load_schemas()


class SlybotJsonSchemaValidator(Draft3Validator):
    DEFAULT_TYPES = Draft3Validator.DEFAULT_TYPES.copy()
    DEFAULT_TYPES.update({
        "mapping": dict,
    })

def is_valid_ipv6_address(address):
    try:
        socket.inet_pton(socket.AF_INET6, address)
    except socket.error:  # not a valid address
        return False
    return True

def get_url_re():
    """
    Generate a URI validation regexp.
    Slightly modified from Django's URL validation
    https://github.com/django/django/blob/master/django/core/validators.py
    """
    ul = '\u00a1-\uffff'  # unicode letters range (must be a unicode string, not a raw string)
    # IP patterns
    ipv4_re = r'(?:25[0-5]|2[0-4]\d|[0-1]?\d?\d)(?:\.(?:25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}'
    ipv6_re = r'\[[0-9a-f:\.]+\]' # Simple RE, validated later

    # Host patterns
    hostname_re = r'[a-z' + ul + r'0-9](?:[a-z' + ul + r'0-9-]{0,61}[a-z' + ul + r'0-9])?'
    # Max length for domain name labels is 63 characters per RFC 1034 sec. 3.1
    domain_re = r'(?:\.(?!-)[a-z' + ul + r'0-9-]{1,63}(?<!-))*'
    tld_re = r'\.(?:[a-z' + ul + r']{2,63}|xn--[a-z0-9]{1,59})\.?'
    host_re = '(' + hostname_re + domain_re + tld_re + '|localhost)'

    return re.compile(
        r'^(?:https?)://'  # Only http or https links allowed
        r'(?:\S+(?::\S*)?@)?'  # user:pass authentication
        r'(?:' + ipv4_re + '|' + ipv6_re + '|' + host_re + ')'
        r'(?::\d{2,5})?'  # port
        r'(?:[/?#][^\s]*)?'  # resource path
        r'\Z',
        re.IGNORECASE
    )

URL_RE = get_url_re()

def get_schema_validator(schema):
    resolver = RefResolver("", schema, _SCHEMAS)

    @FormatChecker.cls_checks('url', (ValueError, UnicodeError))
    def is_valid_uri(url):
        if not isinstance(url, six.string_types):
            return False
        if isinstance(url, six.binary_type):
            url = url.decode('utf-8')

        scheme, netloc, path, query, fragment = urlsplit(url)
        netloc = netloc.encode('idna').decode('ascii')  # IDN -> ACE
        url = urlunsplit((scheme, netloc, path, query, fragment))

        if not URL_RE.match(url):
            return False

        # Validate IPv6
        ipv6_match = re.search(r'^\[(.+)\](?::\d{2,5})?$', netloc)
        if ipv6_match:
            potential_ip = ipv6_match.groups()[0]
            if not is_valid_ipv6_address(potential_ip):
                return False
        return True

    # Workaround for https://github.com/Julian/jsonschema/pull/272
    @FormatChecker.cls_checks('regex', (Exception))
    def is_valid_re(re_source):
        if not isinstance(re_source, six.string_types):
            return False
        if isinstance(re_source, six.binary_type):
            re_source = re_source.decode('utf-8')

        re.compile(re_source)
        return True

    return SlybotJsonSchemaValidator(_SCHEMAS[schema], resolver=resolver,
                                     format_checker=FormatChecker())


def validate_project_schema(specs):

    project = specs["project"]
    get_schema_validator("project").validate(project)

    items = specs["items"]
    get_schema_validator("items").validate(items)

    extractors = specs["extractors"]
    get_schema_validator("extractors").validate(extractors)

    spider_schema_validator = get_schema_validator("spider")
    for spider in specs["spiders"].values():
        spider_schema_validator.validate(spider)

    return True



"""
Tests for apply_annotations
"""

from unittest import TestCase
from slybot.baseurl import insert_base_url, get_base_url
from scrapely.htmlpage import HtmlPage

class TestApplyAnnotations(TestCase):
    def test_insert_base_relative(self):
        """Replace relative base href"""
        html_in = '<html><head><base href="products/"><body></body></html>'
        html_target = '<html><head><base href="http://localhost:8000/products/" />\
<body></body></html>'
        html_out = insert_base_url(html_in, "http://localhost:8000/")
        self.assertEqual(html_out, html_target)

    def test_insert_base_noreplace(self):
        """base tag dont need to be replaced"""
        html_in = html_target = '<html><head><base href="http://localhost:8000/products/"><body></body></html>'
        html_out = insert_base_url(html_in, "http://localhost:8000/users/blog.html")
        self.assertEqual(html_out, html_target)
        
    def test_insert_base_addbase(self):
        """add base tag when not present"""
        html_in = '<html><head><meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1">\
<body></body></html>'
        html_target = '<html><head><base href="http://localhost:8000/" />\
<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1">\
<body></body></html>'
        html_out = insert_base_url(html_in, "http://localhost:8000/")
        self.assertEqual(html_out, html_target)

    def test_insert_base_commented(self):
        """Test weird case when base tag is commented in origin"""
        html_in = '<html><head><!-- <base href="http://example.com/"> --></head>\
<body>Body</body></html>'
        html_target = '<html><head><base href="http://example.com/" />\
<!-- <base href="http://example.com/"> --></head><body>Body</body></html>'
        html_out = insert_base_url(html_in, "http://example.com/")
        self.assertEqual(html_out, html_target)

    def test_insert_base_nohead(self):
        """Test base insert when no head element is present"""
        html_in = '<html><body>Body</body></html>'
        html_target = '<html>\n\
<head><base href="http://localhost:8000/" /></head>\n\
<body>Body</body></html>'
        html_out = insert_base_url(html_in, "http://localhost:8000/")
        self.assertEqual(html_out, html_target)

    def test_get_base_url(self):
        """Basic get_base_url test"""
        html = u'<html><head><base href="http://example.com/products/" />\
<body></body></html>'
        page = HtmlPage("http://example.com/products/p19.html", body=html)
        self.assertEqual(get_base_url(page), "http://example.com/products/")

    def test_get_base_url_nobase(self):
        """Base tag does not exists"""
        html = u'<html><head><body></body></html>'
        page = HtmlPage("http://example.com/products/p19.html", body=html)
        self.assertEqual(get_base_url(page), "http://example.com/products/p19.html")

    def test_get_base_url_empty_basehref(self):
        """Base tag exists but href is empty"""
        html = u'<html><head><base href="" />\
<body></body></html>'
        url = "http://example.com/products/p19.html"
        page = HtmlPage(url, body=html)
        self.assertEqual(get_base_url(page), url)





"""
html page utils
"""
from __future__ import absolute_import

from scrapy.utils.spider import arg_to_iter
from slybot.fieldtypes import FieldTypeManager


_DEFAULT_EXTRACTOR = FieldTypeManager().type_processor_class(u'raw html')()


class cached_property(object):
    """
    A property that is only computed once per instance and then replaces itself
    with an ordinary attribute. Deleting the attribute resets the property.
    Source: https://github.com/bottlepy/bottle/blob/18ea724b6f658943606237e01febc242f7a56260/bottle.py#L162-L173
    """

    def __init__(self, func):
        self.__doc__ = getattr(func, u'__doc__')
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value


def region_id(region, attribute_name='data-tagid'):
    try:
        return region.attributes.get(attribute_name, -1)
    except AttributeError:
        return -1


def load_annotations(extractor):
    for e in arg_to_iter(extractor):
        if hasattr(e, 'annotation') and not hasattr(e, u'extractors'):
            meta = e.annotation.metadata
            if u'attribute' not in meta:
                attribute = [a for a in e.annotation.tag_attributes]
                content = meta.get(u'text-content', u'content')
                attribute, ann = (attribute or [(content, None)])[0]
                if not e.annotation.surrounds_attribute:
                    meta['text-content'] = '#portia-content'
                meta[u'attribute'] = attribute
                if ann is not None:
                    if isinstance(ann, list):
                        ann = ann[0].get(u'field')
                    meta[u'field'] = ann
            if not meta.get(u'field'):
                attr = e.annotation.surrounds_attribute
                if isinstance(attr, list):
                    attr = attr[0].get(u'field')
                meta[u'field'] = attr
            yield e.annotation
        if hasattr(e, u'extractors') and not hasattr(e, u'schema'):
            for sub_e in load_annotations(e.extractors):
                yield sub_e



"""
Numeric data extraction
"""

from scrapely.extractors import contains_any_numbers, extract_number

class NumberTypeProcessor(object):
    """NumberTypeProcessor

    Extracts a number from text

    >>> from scrapely.extractors import htmlregion
    >>> n = NumberTypeProcessor()
    >>> n.extract(htmlregion(u"there are no numbers here"))
    >>> n.extract(htmlregion(u"foo 34"))
    u'foo 34'
    >>> n.adapt(u"foo 34", None)
    u'34'

    If more than one number is present, nothing is extracted
    >>> n.adapt(u"34 48", None) is None
    True
    """
    name = 'number'
    description = 'extracts a single number in the text passed'
    
    def extract(self, htmlregion):
        """Only matches and extracts strings with at least one number"""
        return contains_any_numbers(htmlregion.text_content)
        
    def adapt(self, text, htmlpage=None):
        return extract_number(text)




import json
import logging
import os

from scrapely.htmlpage import HtmlPage
from scrapy import signals
from scrapy.exceptions import NotConfigured
from scrapy.utils.request import request_fingerprint
from scrapy.utils.project import data_path
from slybot.plugins.scrapely_annotations.annotations import _CLUSTER_NA


class PersistentClusteringMiddleware(object):
    def __init__(self, directory, reset=False, stats=None):
        try:
            import anydbm as _dbm
        except ImportError:
            import dbm as _dbm
        self.dbmodule = _dbm
        self.directory = directory
        self.reset = reset
        self.stats = stats
        self.clustering_enabled = False

    @classmethod
    def from_crawler(cls, crawler):
        s = crawler.settings
        if not s.getbool('PERSISTENT_PAGE_CLUSTERING'):
            raise NotConfigured
        directory = data_path(s.get('CLUSTERING_DIR', 'clustering'))
        reset = s.getbool('CLUSTERING_RESET')
        o = cls(directory, reset, crawler.stats)
        crawler.signals.connect(o.spider_opened, signal=signals.spider_opened)
        crawler.signals.connect(o.spider_closed, signal=signals.spider_closed)
        return o

    def spider_opened(self, spider):
        try:
            clustering = spider.plugins['Annotations'].clustering
            assert bool(clustering) == True
            self.clustering_enabled = True
        except (KeyError, AttributeError, AssertionError):
            logging.warning('Persistent page clustering has not been enabled '
                            'because page clustering is not enabled for this '
                            'spider')
            return
        if not os.path.exists(self.directory):
            os.makedirs(self.directory)
        dbpath = os.path.join(self.directory, spider.name)
        flag = 'n' if self.reset else 'c'
        self.db = self.dbmodule.open(dbpath, flag=flag)
        for data in getattr(self.db, 'itervalues', self.db.values)():
            page, encoding = json.loads(data)
            clustering.add_page(HtmlPage(body=page.decode(encoding)))

    def spider_closed(self, spider):
        self.db.close()

    def process_spider_output(self, response, result, spider):
        """Store page tag details if page clustering was not available."""
        saved = False
        for out in result:
            if (not saved and self.clustering_enabled and
                    hasattr(out, 'get') and
                    out.get('_template_cluster') == _CLUSTER_NA):
                key = request_fingerprint(response.request)
                logging.debug('Saving %s for clustering on next crawl',
                              response.request)
                self.db[key] = json.dumps([response.body_as_unicode(),
                                           response.encoding])
                saved = True
            yield out



from datetime import datetime
from itertools import product

import six


class FragmentGenerator(object):
    def _process_fixed(self, fragment):
        return [fragment]

    def _process_list(self, fragment):
        return fragment.split(' ')

    def _process_date(self, fragment):
        now = datetime.now()
        return [now.strftime(fragment)]

    def _process_range(self, fragment):
        a, b = fragment.split('-')

        if a.isalpha() and b.isalpha():
            a, b = [ord(w.lower()) for w in [a, b]]
            return (chr(w) for w in six.moves.range(a, b + 1))
        else:
            a, b = int(a), int(b)
            return (str(i) for i in six.moves.range(a, b + 1))

    def _process_fragment(self, fragment):
        processor = getattr(self, '_process_{}'.format(fragment['type']))
        return processor(fragment['value'])

    def process_fragments(self, spec):
        return map(self._process_fragment, spec['fragments'])

    def __call__(self, spec):
        generated = product(*self.process_fragments(spec))
        for fragment_list in generated:
            yield ''.join(fragment_list)



"""
Field Types

Spiders extracts items of a given type. These item types are defined by a
schema, which specifies the type of each field in the item. This module
contains FieldProcessor implementations, which are the classes responsible for
custom processing of these types.

We keep the types of scrapers supported flexible and allow different methods
for each. In the future, we expect many different types, for example one might
be a mechanical turk scraper and the fields would have to do user validation
and provide error messages.
"""
from .text import (
    RawFieldTypeProcessor,
    TextFieldTypeProcessor,
    SafeHtmlFieldTypeProcessor
)
from .images import ImagesFieldTypeProcessor
from .url import UrlFieldTypeProcessor
from .number import NumberTypeProcessor
from .point import GeoPointFieldTypeProcessor
from .price import PriceTypeProcessor
from .date import DateTimeFieldTypeProcessor


class FieldTypeManager(object):
    _TYPEMAP = dict((c.name, c) for c in (
        RawFieldTypeProcessor, TextFieldTypeProcessor,
        ImagesFieldTypeProcessor, NumberTypeProcessor,
        UrlFieldTypeProcessor, SafeHtmlFieldTypeProcessor,
        GeoPointFieldTypeProcessor, PriceTypeProcessor,
        DateTimeFieldTypeProcessor
    ))
    _names = sorted(_TYPEMAP.keys())

    def available_type_names(self):
        """Find the names of all field types available. """
        return self._names

    def type_processor_class(self, name):
        """Retrieve the class for the given extractor

        This can be useful to introspect on the constructor arguments. If no
        suitable type is found, it will default to the RawFieldTypeProcessor
        (no processing of extracted data is done).
        """
        return self._TYPEMAP.get(name, RawFieldTypeProcessor)

    def all_processor_classes(self):
        """Retrieve all processor classes registered"""
        return list(self._TYPEMAP.values())



from unittest import TestCase

from datetime import datetime

from scrapely.htmlpage import HtmlPage
from scrapely.extraction import InstanceBasedLearningExtractor

from slybot.extractors import (create_regex_extractor, apply_extractors,
                               add_extractors_to_descriptors)
from slybot.fieldtypes import TextFieldTypeProcessor
from slybot.item import create_slybot_item_descriptor
from slybot.plugins.scrapely_annotations.extraction import SlybotIBLExtractor
from slybot.plugins.scrapely_annotations.builder import (
    apply_annotations, _clean_annotation_data
)


class ExtractorTest(TestCase):

    annotated = u"""
<table>
<tr data-scrapy-annotate="{&quot;required&quot;: [], &quot;variant&quot;: 0, &quot;annotations&quot;: {&quot;content&quot;: &quot;gender&quot;}}">
<th class="item-key">Gender</th>
<td >Male</td></tr>
</table>"""
    _target = u"""
<table>
<tr>
<th class="item-key">Gender</th>
<td >Male</td></tr>
</table>"""
    annotated2 = u"""
<table>
<tr data-scrapy-annotate="{&quot;required&quot;: [], &quot;variant&quot;: 0, &quot;annotations&quot;: {&quot;content&quot;: &quot;name&quot;}}">
<th class="item-key">Name</th>
<td >John</td></tr>
<span data-scrapy-annotate="{&quot;required&quot;: [], &quot;variant&quot;: 0, &quot;annotations&quot;: {&quot;content&quot;: &quot;gender&quot;}}">Male</span>
</table>"""
    _target2 = u"""
<body>
<tr>
<th class="item-key">Name</th><td>Olivia</td></tr>
<span></span>
</body>"""

    annotations = _clean_annotation_data([{
        'id': 'annotation1',
        'selector': 'td > a',
        'container_id': 'parent',
        'data': {
            1: {
                'attribute': 'content',
                'field': 'title',
                'required': False,
                'extractors': []
            },
            2: {
                'attribute': 'content',
                'field': 'name',
                'required': False,
                'extractors': ['3']
            },
            3: {
                'attribute': 'href',
                'field': 'url',
                'required': False,
                'extractors': ['1', '2']
            }
        }
    }, {
        'id': 'annotation2',
        'selector': 'span',
        'container_id': 'parent',
        'data': {
            1: {
                'attribute': 'content',
                'field': 'price',
                'required': False,
                'extractors': ['8', '4', '5', '6']
            },
            2: {
                'attribute': 'content',
                'field': 'date',
                'required': False,
                'extractors': ['4', '7']
            }
        }
    }, {
        'id': 'parent',
        'item_container': True,
        'selector': 'body'
    }])
    target3 = u"""
    <html>
    <body>
    <tr>
        <th class="item-key">Name</th>
        <td>
            <a href="/olivia.html">Name: Olivia</a>
        </td>
    </tr><span>2016-03-17 20:25</span>
    </body></html>"""

    template = HtmlPage(url="http://www.test.com/", body=annotated)
    target = HtmlPage(url="http://www.test.com/", body=_target)
    template2 = HtmlPage(url="http://www.test.com/", body=annotated2)
    target2 = HtmlPage(url="http://www.test.com/a", body=_target2)
    template3 = HtmlPage(url="http://www.test.com/a",
                         body=apply_annotations(annotations, target3))
    target3 = HtmlPage(url="http://www.test.com/a", body=target3)

    def test_regex_extractor(self):
        extractor = create_regex_extractor("(\d+).*(\.\d+)")
        extracted = extractor(u"The price of this product is <div>45</div> </div class='small'>.50</div> pounds")
        self.assertEqual(extracted, u"45.50")
        processor = TextFieldTypeProcessor()
        self.assertEqual(processor.adapt(extracted, None), u"45.50")

    def test_raw_type_w_regex(self):
        schema = {
            'fields': {
                'gender': {
                    'required': False,
                    'type': 'raw',
                    'vary': False,
                }
            }
        }
        descriptor = create_slybot_item_descriptor(schema)
        extractors = {
            1: {"regular_expression": "Gender.*(<td\s*>(?:Male|Female)</td>)"}
        }
        apply_extractors(descriptor, {"gender": [1]}, extractors)

        ibl_extractor = SlybotIBLExtractor([
            (self.template, {'#default': descriptor}, '0.12.0')])
        self.assertEqual(ibl_extractor.extract(self.target)[0][0].dump()['gender'], [u'<td >Male</td>'])

    def test_negative_hit_w_regex(self):
        schema = {
            'fields': {
                'gender': {
                    'required': False,
                    'type': 'number',
                    'vary': False,
                }
            }
        }
        descriptor = create_slybot_item_descriptor(schema)
        extractors = {1: {"regular_expression": "Gender\\s+(Male|Female)"}}
        apply_extractors(descriptor, {"gender": [1]}, extractors)

        ibl_extractor = SlybotIBLExtractor([
            (self.template, {'#default': descriptor}, '0.12.0')])
        self.assertEqual(ibl_extractor.extract(self.target)[0][0].dump(), {})

    def test_text_type_w_regex(self):
        schema = {
            "fields": {
                'gender': {
                    'required': False,
                    'type': 'text',
                    'vary': False,
                }
            }
        }
        descriptor = create_slybot_item_descriptor(schema)
        extractors = {1: {"regular_expression": "Gender\\s+(Male|Female)"}}
        apply_extractors(descriptor, {"gender": [1]}, extractors)

        ibl_extractor = SlybotIBLExtractor([
            (self.template, {'#default': descriptor}, '0.12.0')])
        self.assertEqual(ibl_extractor.extract(self.target)[0][0].dump()['gender'], [u'Male'])

    def test_type_extractor(self):
        schema = {
            "fields": {
                'gender': {
                    'required': False,
                    'type': 'number',
                    'vary': False,
                }
            }
        }
        descriptor = create_slybot_item_descriptor(schema)
        extractors = {
            1: {"type_extractor": "text"},
            2: {"regular_expression": "Gender\\s+(Male|Female)"}
        }
        apply_extractors(descriptor, {"gender": [1, 2]}, extractors)

        ibl_extractor = SlybotIBLExtractor([
            (self.template, {'#default': descriptor}, '0.12.0')])
        self.assertEqual(ibl_extractor.extract(self.target)[0][0].dump()['gender'], [u'Male'])

    def test_default_type_extractor(self):
        schema = {
            'fields': {}
        }
        descriptor = create_slybot_item_descriptor(schema)
        extractors = {
            1: {"regular_expression": "Gender\\s+(Male|Female)"}
        }
        apply_extractors(descriptor, {"gender": [1]}, extractors)

        ibl_extractor = SlybotIBLExtractor([
            (self.template, {'#default': descriptor}, '0.12.0')])
        self.assertEqual(ibl_extractor.extract(self.target)[0][0].dump()['gender'], [u'Male'])

    def test_text_type_w_regex_and_no_groups(self):
        schema = {
            'fields': {
                'gender': {
                    'required': False,
                    'type': 'text',
                    'vary': False,
                }
            }
        }
        descriptor = create_slybot_item_descriptor(schema)
        extractors = {
            1: {"regular_expression": "Gender"}
        }
        apply_extractors(descriptor, {"gender": [1]}, extractors)

        ibl_extractor = SlybotIBLExtractor([
            (self.template, {'#default': descriptor}, '0.12.0')])
        self.assertEqual(ibl_extractor.extract(self.target)[0][0].dump()['gender'], [u'Gender'])

    def test_extractor_w_empty_string_extraction(self):
        schema = {
            'fields': {
                'gender': {
                    'required': False,
                    'type': 'text',
                    'vary': False,
                },
                'name': {
                    'required': True,
                    'type': 'text',
                    'vary': False,
                }
            }
        }
        descriptor = create_slybot_item_descriptor(schema)
        extractors = {
            1: {
                "regular_expression": "([0-9]+)"
            }
        }
        apply_extractors(descriptor, {"gender": [1]}, extractors)

        ibl_extractor = SlybotIBLExtractor([
            (self.template2, {'#default': descriptor}, '0.12.0')])
        self.assertEqual(ibl_extractor.extract(self.target2)[0][0].dump()['name'], [u'Name Olivia'])

    def test_per_annotation_extractors(self):
        schema = {
            'fields': {
                'url': {
                    'required': False,
                    'type': 'text',
                    'vary': False,
                },
                'name': {
                    'required': True,
                    'type': 'text',
                    'vary': False,
                }
            }
        }
        extractors = {
            '1': {
                'type_extractor': 'url'
            },
            '2': {
                'regular_expression': '(.*)\.html'
            },
            '3': {
                'regular_expression': 'Name: (.*)'
            },
            '4': {
                'type_extractor': 'text'
            },
            '5': {
                'type_extractor': 'price'
            },
            '6': {
                'type_extractor': 'number'
            },
            '7': {
                'type_extractor': 'date'
            },
            '8': {
                'regular_expression': '(\d+)-'
            }
        }
        descriptors = {'#default': create_slybot_item_descriptor(schema)}
        add_extractors_to_descriptors(descriptors, extractors)
        ibl_extractor = SlybotIBLExtractor([
            (self.template3, descriptors, '0.13.0')
        ])
        result = {u'_template': '6223d000057491040e4f411cf1f0734ea802eeb6',
                  'name': [u'Olivia'], 'url': [u'http://www.test.com/olivia'],
                  'title': [u'Name: Olivia'], 'price': [u'2016'],
                  'date': [datetime(2016, 3, 17, 20, 25)]}
        data = ibl_extractor.extract(self.target3)[0][0].dump()
        self.assertEqual(data, result)



import re
import itertools
from lxml import html

from scrapy.http.request.form import _get_inputs


class GenericForm:

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def _pick_node(self, doc, selector):
        nodes = doc.xpath(selector['xpath'])
        if nodes:
            return nodes[0]

    def _filter_by_regex(self, lines, regex):
        search_regex = re.compile(regex).search
        return [l for l in lines if search_regex(l)]

    def _get_field_values(self, form, field_descriptor):
        if 'name' in field_descriptor:
            field_name = field_descriptor['name']
        else:
            select_field = self._pick_node(form, field_descriptor)
            field_name = select_field.name

        field_type = field_descriptor['type']
        if field_type == 'constants':
            return [[field_name, option] for option in self.get_value(field_descriptor)]
        elif field_type == 'iterate':
            select_field = self._pick_node(form, field_descriptor)
            values = self._filter_by_regex(select_field.value_options,
                                           self.get_value(field_descriptor))
            return [[select_field.name, option] for option in values]
        elif field_type == 'inurl':
            return [[field_name, option] for option in field_descriptor['file_values']]

    def get_value(self, field_descriptor):
        values = field_descriptor.get('value', '')
        if isinstance(values, list):
            return [val.format(**self.kwargs) for val in values]
        else:
            return values.format(**self.kwargs)

    def set_values_url_field(self, field_descriptor, body):
        field_descriptor['file_values'] = body.split('\n')

    def get_url_field(self, form_descriptor):
        for i, field_descriptor in enumerate(form_descriptor['fields']):
            if (field_descriptor['type'] == 'inurl'
                and (not 'file_values' in field_descriptor or
                     not field_descriptor['file_values'])):
                yield i, field_descriptor

    def fill_generic_form(self, url, body, form_descriptor):

        doc = html.document_fromstring(body, base_url=url)
        form = self._pick_node(doc, form_descriptor)
        if form is None:
            raise Exception('Generic form not found')

        # Get all the possible inputs for each field
        values = [self._get_field_values(form, field)
                  for field in form_descriptor['fields']]

        for params in itertools.product(*values):
            form_values = dict(_get_inputs(form, None, False, None, None))
            for name, option in params:
                form_values[name] = option
            yield list(form_values.items()), form.action or form.base_url, form.method



from six.moves.urllib_parse import urlparse
import os
import json
import re

from collections import OrderedDict

from scrapely.htmlpage import HtmlPage, HtmlTag, HtmlTagType
from scrapy.utils.misc import load_object


TAGID = u"data-tagid"
GENERATEDTAGID = u"data-genid"
OPEN_TAG = HtmlTagType.OPEN_TAG
CLOSE_TAG = HtmlTagType.CLOSE_TAG
UNPAIRED_TAG = HtmlTagType.UNPAIRED_TAG


def iter_unique_scheme_hostname(urls):
    """Return an iterator of tuples (scheme, hostname) over the given urls,
    filtering dupes
    """
    scheme_hostname = set()
    for x in urls:
        p = urlparse(x)
        scheme_hostname.add((p.scheme, p.hostname))
    return list(scheme_hostname)


def open_project_from_dir(project_dir):
    specs = {"spiders": {}}
    try:
        with open(os.path.join(project_dir, "project.json")) as f:
            specs["project"] = json.load(f)
    except IOError:
        specs["project"] = {}
    with open(os.path.join(project_dir, "items.json")) as f:
        specs["items"] = json.load(f)
    with open(os.path.join(project_dir, "extractors.json")) as f:
        specs["extractors"] = json.load(f)

    spec_base = os.path.join(project_dir, "spiders")
    for fname in os.listdir(spec_base):
        if fname.endswith(".json"):
            spider_name = os.path.splitext(fname)[0]
            with open(os.path.join(spec_base, fname)) as f:
                try:
                    spec = json.load(f)
                    template_names = spec.get("template_names")
                    if template_names:
                        templates = load_external_templates(spec_base,
                                                            spider_name,
                                                            template_names)
                        spec.setdefault("templates", []).extend(templates)
                    else:
                        templates = []
                        for template in spec.get('templates', []):
                            if template.get('version') < '0.13.0':
                                templates.append(template)
                            else:
                                templates.append(_build_sample(template))
                    specs["spiders"][spider_name] = spec
                except ValueError as e:
                    raise ValueError(
                        "Error parsing spider (invalid JSON): %s: %s" %
                        (fname, e)
                    )
    return specs


def load_external_templates(spec_base, spider_name, template_names):
    """A generator yielding the content of all passed `template_names` for
    `spider_name`.
    """
    for name in template_names:
        with open(os.path.join(spec_base, spider_name, name + ".json")) as f:
            sample = json.load(f)
            samples_sub_dir = os.path.join(spec_base, spider_name, name)
            if (os.path.exists(samples_sub_dir) and
                    os.path.isdir(samples_sub_dir)):
                for fname in os.listdir(samples_sub_dir):
                    if fname.endswith('.html'):
                        with open(os.path.join(samples_sub_dir, fname)) as f:
                            attr = fname[:-len('.html')]
                            sample[attr] = f.read().decode('utf-8')
            yield _build_sample(sample)


def _build_sample(sample):
    from slybot.plugins.scrapely_annotations.builder import Annotations
    data = sample.get('plugins', {}).get('annotations-plugin')
    if data:
        Annotations().save_extraction_data(data, sample)
    sample['page_id'] = sample.get('page_id') or sample.get('id') or ""
    return sample


def htmlpage_from_response(response, _add_tagids=False):
    body = response.body_as_unicode()
    if _add_tagids:
        body = add_tagids(body)
    return HtmlPage(response.url, response.headers, body,
                    encoding=response.encoding)


def load_plugins(settings):
    if settings.get('LOADED_PLUGINS', None):
        return settings.get('LOADED_PLUGINS', None)
    plugins = settings['PLUGINS']
    if plugins:
        return [load_object(p) if isinstance(p, str) else p for p in plugins]
    else:
        from slybot.plugins.scrapely_annotations import Annotations
        return [Annotations]


def load_plugin_names(settings):
    """
    Generate a unique name for a plugin based on the class name module name
    and path

    >>> settings = {'PLUGINS': ['a', 'b.c', 'a.c']}
    >>> load_plugin_names(settings)
    ['a', 'c', 'a.c']
    """
    seen = set()

    def generate_name(path, maxsplit=0, splits=None):
        if splits is None:
            splits = len(path.split('.')) - 1
        name = '.'.join(path.split('.', splits - maxsplit)[-1].rsplit('.',
                        maxsplit))
        if name not in seen or maxsplit >= splits:
            seen.add(name)
            return name
        return generate_name(path, maxsplit + 1, splits)

    if settings['PLUGINS']:
        return [generate_name(path) for path in settings['PLUGINS']]
    else:
        return ['Annotations']


def include_exclude_filter(include_patterns, exclude_patterns):
    filterf = None
    includef = None
    if include_patterns:
        pattern = include_patterns[0] if len(include_patterns) == 1 else \
            "(?:%s)" % '|'.join(include_patterns)
        includef = re.compile(pattern).search
        filterf = includef
    if exclude_patterns:
        pattern = exclude_patterns[0] if len(exclude_patterns) == 1 else \
            "(?:%s)" % '|'.join(exclude_patterns)
        excludef = re.compile(pattern).search
        if not includef:
            filterf = lambda x: not excludef(x)
        else:
            filterf = lambda x: includef(x) and not excludef(x)
    return filterf if filterf else bool


class IndexedDict(OrderedDict):
    """
    Ordered dictionary where values can also be obtained by their index as if
    they were in a list

    >>> idd = IndexedDict([('spam', 1), ('eggs', 2), ('bacon', 3)])
    >>> idd['spam']
    1
    >>> idd[0]
    1
    >>> idd['bacon']
    3
    >>> idd[2]
    3
    >>> idd[2] = 'ham'
    Traceback (most recent call last):
        ...
    TypeError: keys must not be an integers
    >>> idd[3]
    Traceback (most recent call last):
        ...
    IndexError: index out of range
    """
    def __setitem__(self, key, value):
        if isinstance(key, int):
            raise TypeError("keys must not be an integers")
        super(IndexedDict, self).__setitem__(key, value)

    def __getitem__(self, key):
        if isinstance(key, int):
            if key >= len(self):
                raise IndexError('index out of range')
            for i, k in enumerate(self):
                if i == key:
                    key = k
                    break
        return super(IndexedDict, self).__getitem__(key)


def _quotify(mystr):
    """
    quotifies an html tag attribute value.
    Assumes then, that any ocurrence of ' or " in the
    string is escaped if original string was quoted
    with it.
    So this function does not altere the original string
    except for quotation at both ends, and is limited just
    to guess if string must be quoted with '"' or "'"
    """
    quote = '"'
    l = len(mystr)
    for i in range(l):
        if mystr[i] == "\\" and i + 1 < l and mystr[i + 1] == "'":
            quote = "'"
            break
        elif mystr[i] == "\\" and i + 1 < l and mystr[i + 1] == '"':
            quote = '"'
            break
        elif mystr[i] == "'":
            quote = '"'
            break
        elif mystr[i] == '"':
            quote = "'"
            break
    return quote + mystr + quote


def serialize_tag(tag):
    """
    Converts a tag into a string when a slice [tag.start:tag.end]
    over the source can't be used because tag has been modified
    """
    out = "<"
    if tag.tag_type == HtmlTagType.CLOSE_TAG:
        out += "/"
    out += tag.tag

    attributes = []
    for key, val in tag.attributes.items():
        aout = key
        if val is not None:
            aout += "=" + _quotify(val)
        attributes.append(aout)
    if attributes:
        out += " " + " ".join(attributes)

    if tag.tag_type == HtmlTagType.UNPAIRED_TAG:
        out += "/"
    return out + ">"


def _must_add_tagid(element):
    return (isinstance(element, HtmlTag) and
            element.tag_type != CLOSE_TAG and
            element.tag != 'ins')


def _modify_tagids(source, add=True):
    """Add or remove tags ids to/from HTML document"""
    output = []
    tagcount = 0
    if not isinstance(source, HtmlPage):
        source = HtmlPage(body=source)
    for element in source.parsed_body:
        if _must_add_tagid(element):
            if add:
                element.attributes[TAGID] = str(tagcount)
                tagcount += 1
            else:  # Remove previously added tagid
                element.attributes.pop(TAGID, None)
            output.append(serialize_tag(element))
        else:
            output.append(source.body[element.start:element.end])
    return ''.join(output)


def add_tagids(source):
    """
    Applies a unique attribute code number for each tag element in order to be
    identified later in the process of apply annotation"""
    return _modify_tagids(source)


def remove_tagids(source):
    """remove from the given page, all tagids previously added by add_tagids()
    """
    return _modify_tagids(source, False)



# -*- coding: utf-8 -*-
import re

from unittest import TestCase
from os.path import dirname, join

from slybot.validation.schema import get_schema_validator, \
            ValidationError, validate_project_schema
from slybot.utils import open_project_from_dir

_TEST_PROJECT_DIR = join(dirname(__file__), "data/SampleProject")

def spider_json(start_urls):
    return {
        "start_urls": start_urls,
        "links_to_follow": "none",
        "respect_nofollow": True,
        "templates": [],
    }

def start_url_schema(url_string):
    return {"url": url_string, "type": "url"}

class JsonSchemaTest(TestCase):

    def test_regex_formatting_wrong(self):
        obj = {
            "0": {
                "regular_expression": "Item: (\d+"
            }
        }
        validator = get_schema_validator("extractors")
        with self.assertRaises(ValidationError):
            validator.validate(obj)

    def test_regex_formatting_ok(self):
        obj = {
            "0": {
                "regular_expression": "Item: (\d+)"
            }
        }
        validator = get_schema_validator("extractors")
        self.assertEqual(validator.validate(obj), None)

    def test_valid_url(self):
        legacy_start_urls = [
            'http://www.example.com/',
            'http://www.example.com/經濟',
            'http://www.example.com/?q=經濟',
            'http://www.example.com/#經濟',
            'http://faß.de',
            'http://例.jp/',
            'http://[2001:0000:1234:0000:0000:C1C0:ABCD:0876]/foo/bar',
            'http://[2001::]/foo/bar',
            'http://8.8.8.8/foo/bar',
            'http://localhost:8080/foo/bar',
            'http://foo@localhost:8080/foo/bar',
            'http://foo:bar@localhost:8080/foo/bar',
            'http://domain.com/path/file.html?param=FOO^111&param2=bar&param3=true&_param4=on', # Anonymized URL form sentry d46840d2457c4042b1b58f2fa40e984b
            'https://domain.com/path/file.htm?param=foo#hash/foo/bar/baz:foo|bar:baz',          # Anonymized URL from sentry 01dd2fa09d9540b69ebd33372b2b3a2d
            'https://domain.com/path/file.htm?param=foo#hash/foo/bar/baz:foo|bar%5B%5D:12345',  # Anonymized URL from sentry 87d49ee751494c90a8941dcbdacea634
            'http://domain.com/path?bar[foo]=baz&foo[bar]=12345',                               # Anonymized URL from sentry 9f6835f5decd4d57b9475f04f0a58bd4
        ]

        start_urls = map(start_url_schema, legacy_start_urls)
        validator = get_schema_validator("spider")

        self.assertEqual(validator.validate(spider_json(legacy_start_urls)), None)
        self.assertEqual(validator.validate(spider_json(start_urls)), None)

    def test_invalid_url(self):
        legacy_start_urls = (
            12345, # Not a string
            'example.com', # Lacks protocol
            'http://[:::1]/foo/bar', # Bad IPv6 addr
            'http://http://foo.com/bar', # Double protocol
            'spotify:foobar', # Not http/s protocol
            '/foo', # relative
            '?foo', # relative
            '#foo', # relative
        )

        validator = get_schema_validator("spider")
        for invalid_url in legacy_start_urls:
            with self.assertRaises(ValidationError):
                validator.validate(spider_json([invalid_url]))

    def test_valid_mixed_fragments(self):
        start_urls = [
            {'type': 'url', 'url': 'http://www.example.com/'},
            {'type': 'generated', 'url': 'http://', 'fragments': [
                {'type': 'fixed', 'value': 'http://'}
            ]},
        ]
        validator = get_schema_validator("spider")

        self.assertEqual(validator.validate(spider_json(start_urls)), None)

    def test_valid_fragments(self):
        fragments = (
            [
                {'type': 'fixed', 'value': 'domain.com'},
                {'type': 'range', 'value': '0-10'}
            ],
            [{'type': 'range', 'value': '0-10'}],
            [{'type': 'list', 'value': 'a b c'}],
            [{'type': 'list', 'value': 'one_element'}]
        )

        validator = get_schema_validator("spider")
        for fragment in fragments:
            start_url = [{
                'url': 'http://domain.com',
                'type': 'generated',
                'fragments': fragment
            }]
            self.assertEqual(validator.validate(spider_json(start_url)), None)

    def test_test_project(self):
        specs = open_project_from_dir(_TEST_PROJECT_DIR)
        self.assertTrue(validate_project_schema(specs))



from .text import TextFieldTypeProcessor
from dateparser.date import DateDataParser


class DateTimeFieldTypeProcessor(TextFieldTypeProcessor):
    """
    Extracts a date from text

    >>> from scrapely.extractors import htmlregion
    >>> d = DateTimeFieldTypeProcessor()
    >>> d.extract(htmlregion(u"  Jan 12, 2014 11:15AM  <br>"))
    u'Jan 12, 2014 11:15AM'
    >>> d.adapt(u"Jan 12, 2014 11:15AM", None).strftime('%Y-%m-%dT%H:%M:%S')
    '2014-01-12T11:15:00'
    >>> d.adapt(u'no date here', None)
    """

    name = 'date'
    description = 'Extracts date and time information from a string'

    def extract(self, htmlregion):
        return super(DateTimeFieldTypeProcessor, self).extract(htmlregion)

    def adapt(self, text, htmlpage=None):
        try:
            return DateDataParser().get_date_data(text)['date_obj']
        except ValueError:
            return



from __future__ import absolute_import
try:
    from cStringIO import StringIO # py2
except ImportError:
    from io import StringIO # py3
from scrapy.link import Link
from .base import BaseLinkExtractor
import csv
import six

# see http://docs.python.org/2/library/csv.html#csv-fmt-params
_FORMAT_PARAMETERS = (
    ('delimiter', ','),
    ('quotechar', '"'),
    ('doublequote', True),
    ('escapechar', None),
    ('lineterminator', '\r\n'),
    ('skipinitialspace', False),
    ('strict', False),
)

class CsvLinkExtractor(BaseLinkExtractor):
    def __init__(self, column=0, **kwargs):
        self.fmtparams = dict((key, kwargs.pop(key, default)) for key, default in _FORMAT_PARAMETERS)
        if six.PY2:
            for key, val in self.fmtparams.items():
                if isinstance(val, unicode):
                    self.fmtparams[key] = val.encode()
        super(CsvLinkExtractor, self).__init__(**kwargs)
        self.allowed_schemes = [x for x in self.allowed_schemes if x and isinstance(x, six.string_types)]
        self.column = column

    def _extract_links(self, response):
        buff = StringIO(response.body)
        reader = csv.reader(buff, **self.fmtparams)
        for row in reader:
            if len(row) > self.column:
                yield Link(row[self.column])




from scrapy.http import Response
from scrapy.link import Link

from page_finder import LinkAnnotation
from .html import HtmlLinkExtractor


class PaginationExtractor(HtmlLinkExtractor):
    def __init__(self, **specs):
        self.link_annotation = LinkAnnotation()
        self.visited = set()
        self.url_to_link = {}
        start_urls = specs.get('start_urls')
        if start_urls:
            self.link_annotation.load(start_urls)
            for url in start_urls:
                self.url_to_link[url] = Link(url)
                self.visited.add(url)
                self.link_annotation.mark_link(url, follow=True)
        super(PaginationExtractor, self).__init__()

    def _extract_links(self, response_or_htmlpage, n_links=3):
        self.visited.add(response_or_htmlpage.url)
        new_links = list(
            super(PaginationExtractor, self)._extract_links(response_or_htmlpage))
        for link in new_links:
            self.url_to_link[link.url] = link
        self.link_annotation.load(link.url for link in new_links)
        if isinstance(response_or_htmlpage, Response):
            n_items = response_or_htmlpage.meta.get('n_items')
        else:
            n_items = response_or_htmlpage.headers.get('n_items')
        if n_items is not None:
            self.link_annotation.mark_link(
                response_or_htmlpage.url, follow=(n_items > 0))
        best = self.link_annotation.best_links_to_follow()
        if best:
            pages = []
            for url in best:
                if url not in self.visited:
                    pages.append(self.url_to_link[url]) # TODO: extract only the best link?
                    if len(pages) == n_links:
                        return pages
        return new_links



from datetime import datetime

from unittest import TestCase

from scrapy.settings import Settings

from slybot.starturls import IdentityGenerator, UrlGenerator
from slybot.validation.schema import get_schema_validator


class StartUrlGenerators(TestCase):
    github_start_urls = [
        'https://github.com/scrapinghub',
        'https://github.com/scrapy',
        'https://github.com/scrapy-plugins'
    ]
    donedeal_start_urls = [
        'https://www.donedeal.ie/cars-for-sale/i/1',
        'https://www.donedeal.ie/cars-for-sale/i/2',
        'https://www.donedeal.ie/houses-for-sale/i/1',
        'https://www.donedeal.ie/houses-for-sale/i/2',
        'https://www.donedeal.ie/pets-for-sale/i/1',
        'https://www.donedeal.ie/pets-for-sale/i/2',
        'https://www.donedeal.ie/kitchens-for-sale/i/1',
        'https://www.donedeal.ie/kitchens-for-sale/i/2'
    ]
    specs = {
        'defaults': [{
            "template": "https://github.com/{}",
            "paths": [{
                "type": "default",
                "values": ["scrapinghub", "scrapy", "scrapy-plugins"],
            }],
            "params": [],
            "params_template": {}
        }],
        'options': [{
            "template": "https://github.com/{}",
            "paths": [{
                "type": "options",
                "values": ["scrapinghub", "scrapy", "scrapy-plugins"],
            }],
            "params": [],
            "params_template": {}
        }],
        'dates': [{
            "template": "http://www.commitstrip.com/{}/{}/{}",
            "paths": [{
                "type": "default",
                "values": ["en"]
            }, {
                "type": "date",
                "values": ["%Y"],
            }, {
                "type": "date",
                "values": ["%m"]
            }],
            "params": [],
            "params_template": {}
        }],
        'range': [{
            "template": "https://www.donedeal.ie/{}/{}/{}",
            "paths": [{
                "type": "default",
                "values": ["cars-for-sale"]
            }, {
                "type": "options",
                "values": ["i"],
            }, {
                "type": "range",
                "values": [100000010, 100000000, -1]
            }],
            "params": [],
            "params_template": {}
        }],
        'params_range': [{
            "template": "http://www.smbc-comics.com/{}",
            "paths": [{
                "type": "default",
                "values": ["index.php"]
            }],
            "params": [{
                "name": "p",
                "type": "range",
                "values": [20, 30, 5]
            }, {
                "name": "q",
                "type": "options",
                "values": ['comic']
            }],
            "params_template": {}
        }],
        'spider_args': [{
            "template": "https://www.donedeal.ie/{}/{}/{}",
            "paths": [{
                "type": "spider_args",
                "values": ["categories", "sections"]
            }, {
                "type": "options",
                "values": ["i"],
            }, {
                "type": "range",
                "values": [1, 3]
            }],
            "params": [],
            "params_template": {}
        }],
        'settings': [{
            "template": "https://www.donedeal.ie/{}/{}/{}",
            "paths": [{
                "type": "settings",
                "values": ["categories", "sections"]
            }, {
                "type": "options",
                "values": ["i"],
            }, {
                "type": "range",
                "values": [1, 3]
            }],
            "params": [],
            "params_template": {}
        }],
        'params': [{
            "template": "https://encrypted.google.com/search",
            "paths": [],
            "params": [{
                "name": "q",
                "type": "options",
                "values": ["nosetests", "tox"]
            }, {
                "name": "location",
                "type": "options",
                "values": ["dublin", "cork"]
            }],
            "params_template": [
                ("hl", "en"),
                ("q", "python unittest")
            ]
        }]
    }

    def test_schema_format(self):
        validator = get_schema_validator('spider')
        spider = {
            'start_urls_type': 'generated_urls',
            'start_urls': [],
            'links_to_follow': 'none',
            'respect_nofollow': True
        }
        for spec in self.specs.values():
            spider['generated_urls'] = spec
            validator.validate(spider)

    def test_start_urls(self):
        self.assertEqual(self.github_start_urls,
                         IdentityGenerator()(self.github_start_urls))

    def test_generate_start_urls_from_defaults(self):
        genny = UrlGenerator()
        spec = self.specs['defaults']
        self.assertEqual(["https://github.com/scrapinghub"],
                         list(genny(spec[0])))

    def test_generate_start_urls_from_options(self):
        genny = UrlGenerator()
        spec = self.specs['options']
        self.assertEqual(self.github_start_urls, list(genny(spec[0])))

    def test_generate_start_urls_from_date(self):
        now = datetime.now()
        genny = UrlGenerator()
        spec = self.specs['dates']
        url = "http://www.commitstrip.com/en/{}/{:02}".format(now.year,
                                                              now.month)
        self.assertEqual([url], list(genny(spec[0])))

    def test_generate_start_urls_from_range(self):
        genny = UrlGenerator()
        spec = self.specs['range']
        urls = ["https://www.donedeal.ie/cars-for-sale/i/%s" % i
                for i in range(100000010, 100000000, -1)]
        self.assertEqual(urls, list(genny(spec[0])))

    def test_generate_start_urls_from_params_range(self):
        genny = UrlGenerator()
        spec = self.specs['params_range']
        urls = ["http://www.smbc-comics.com/index.php?p=%s&q=comic" % i
                for i in range(20, 30, 5)]
        self.assertEqual(urls, list(genny(spec[0])))

    def test_generate_start_urls_from_spider_arg(self):
        genny = UrlGenerator(spider_args={
            'categories': ['cars-for-sale', 'houses-for-sale'],
            'sections': ['pets-for-sale', 'kitchens-for-sale']
        })
        spec = self.specs['spider_args']
        self.assertEqual(self.donedeal_start_urls, list(genny(spec[0])))

    def test_generate_start_urls_from_setting(self):
        genny = UrlGenerator(Settings(values={
            'categories': 'cars-for-sale,houses-for-sale',
            'sections': ['pets-for-sale', 'kitchens-for-sale']
        }))
        spec = self.specs['settings']
        self.assertEqual(self.donedeal_start_urls, list(genny(spec[0])))

    def test_generate_start_urls_from_params(self):
        genny = UrlGenerator()
        spec = self.specs['params']
        base = "https://encrypted.google.com/search?hl=en&q=%s&location=%s"
        n, t, d, c = "nosetests", "tox", "dublin", "cork"
        arg = [(n, d), (n, c), (t, d), (t, c)]
        self.assertEqual([base % (q, l) for q, l in arg], list(genny(spec[0])))

    def test_misconfigured_start_urls_spec_type(self):
        genny = UrlGenerator()
        spec = [{
            "template": "http://www.smbc-comics.com/{}",
            "paths": [{
                "type": "defaults",
                "values": ["index.php"]
            }],
            "params": [],
            "params_template": {}
        }]
        self.assertEqual([], list(genny(spec[0])))

    def test_missing_arg_for_start_urls_spec(self):
        genny = UrlGenerator(Settings(values={'home': 'home.php'}), {
            'index': 'index.php'
        })
        spec = [{
            "template": "http://www.smbc-comics.com/{}",
            "paths": [{
                "type": "spider_args",
                "values": ["home"]
            }],
            "params": [],
            "params_template": {}
        }]
        self.assertEqual([], list(genny(spec[0])))
        spec = [{
            "template": "http://www.smbc-comics.com/{}",
            "paths": [{
                "type": "settings",
                "values": ["index"]
            }],
            "params": [],
            "params_template": {}
        }]
        self.assertEqual([], list(genny(spec[0])))



"""
Link extraction for auto scraping
"""
import re
from six.moves.urllib.parse import urljoin
from scrapy.utils.markup import replace_entities
from scrapy.link import Link
from scrapy.http import HtmlResponse

from scrapely.htmlpage import HtmlTag, HtmlTagType

from slybot.linkextractor.base import BaseLinkExtractor
from slybot.utils import htmlpage_from_response

_META_REFRESH_CONTENT_RE = re.compile(r"(?P<int>(\d*\.)?\d+)\s*;\s*url=(?P<url>.*)")
_ONCLICK_LINK_RE = re.compile("(?P<sep>('|\"))(?P<url>.+?)(?P=sep)")

class HtmlLinkExtractor(BaseLinkExtractor):
    """Link extraction for auto scraping

    Links (urls and the anchor text) are extracted from HtmlPage objects.

    Some safe normalization is done (always correct, does not make assumptions
    about how the site handles URLs). It allows some customization, which we
    expect to learn for specific websites from the crawl logs.
    """

    def _extract_links(self, response_or_htmlpage):
        """Extract links to follow from an html page

        This uses `iterlinks` to read the links in the page.
        """
        htmlpage = htmlpage_from_response(response_or_htmlpage) if \
                    isinstance(response_or_htmlpage, HtmlResponse) else response_or_htmlpage
        return iterlinks(htmlpage)

def iterlinks(htmlpage):
    """Iterate through the links in the HtmlPage passed

    For example:
    >>> from scrapely.htmlpage import HtmlPage
    >>> p = HtmlPage(body=u"Please visit <a href='http://scrapinghub.com/'>Scrapinghub</a>")
    >>> iterlinks(p).next()
    Link(url='http://scrapinghub.com/', text=u'Scrapinghub', fragment='', nofollow=False)
    >>> p = HtmlPage(body=u"Go <a href='home.html'>Home</a>")
    >>> iterlinks(p).next()
    Link(url='home.html', text=u'Home', fragment='', nofollow=False)

    When a url is specified, absolute urls are made:
    >>> p.url = 'http://scrapinghub.com/'
    >>> iterlinks(p).next()
    Link(url='http://scrapinghub.com/home.html', text=u'Home', fragment='', nofollow=False)

    Base href attributes in the page are respected
    >>> p.body = u"<html><head><base href='myproject/'/></head><body>see my <a href='index.html'>project</a></body>"
    >>> iterlinks(p).next()
    Link(url='http://scrapinghub.com/myproject/index.html', text=u'project', fragment='', nofollow=False)
    >>> p.body = u"<html><head><base href='http://scrape.io\\\\' /></head><body>see my <a href='index.html'>project</a></body>"
    >>> iterlinks(p).next()
    Link(url='http://scrape.io/index.html', text=u'project', fragment='', nofollow=False)

    Frameset and iframe urls are extracted
    >>> p = HtmlPage(body=u"<html><frameset><frame src=frame1.html><frame src=frame2.html></frameset><iframe src='iframe.html'/></html>")
    >>> [l.url for l in iterlinks(p)]
    ['frame1.html', 'frame2.html', 'iframe.html']

    As are meta refresh tags:
    >>> p = HtmlPage(body=u"<html><head><meta http-equiv='refresh' content='5;url=http://example.com/' />")
    >>> iterlinks(p).next().url
    'http://example.com/'

    nofollow is set to True if the link has a rel='nofollow' attribute:
    >>> p = HtmlPage(body=u"<a href='somewhere.html' rel='nofollow'>somewhere</a>")
    >>> list(iterlinks(p))
    [Link(url='somewhere.html', text=u'somewhere', fragment='', nofollow=True)]

    It does not require well formed HTML and behaves similar to many browsers
    >>> p = HtmlPage(body=u"<a href='foo'>foo <a href=bar>bar</a><a href='baz'/>baz")
    >>> list(iterlinks(p))
    [Link(url='foo', text=u'foo ', fragment='', nofollow=False), Link(url='bar', text=u'bar', fragment='', nofollow=False), Link(url='baz', text=u'baz', fragment='', nofollow=False)]

    Leading and trailing whitespace should be removed, including in base href
    >>> p = HtmlPage(body=u"<head><base href=' foo/ '/></head><a href='bar '/>baz")
    >>> list(iterlinks(p))
    [Link(url='foo/bar', text=u'baz', fragment='', nofollow=False)]

    Test standard onclick links
    >>> p = HtmlPage(url="http://www.example.com", body=u"<html><td onclick=window.open('page.html?productid=23','win2') >")
    >>> list(iterlinks(p))
    [Link(url='http://www.example.com/page.html?productid=23', text=None, fragment='', nofollow=False)]

    >>> p = HtmlPage("http://www.example.com", body=u"<html><a href='#' onclick=window.open('page.html?productid=24','win2') >")
    >>> list(iterlinks(p))
    [Link(url='http://www.example.com/page.html?productid=24', text=None, fragment='', nofollow=False)]

    >>> p = HtmlPage(body=u"<html><div onclick=window.location.href='http://www.jungleberry.co.uk/Fair-Trade-Earrings/Aguas-Earrings.htm'>")
    >>> list(iterlinks(p))
    [Link(url='http://www.jungleberry.co.uk/Fair-Trade-Earrings/Aguas-Earrings.htm', text=None, fragment='', nofollow=False)]

    Onclick with no href
    >>> p = HtmlPage("http://www.example.com", body=u"<html><a onclick=window.open('page.html?productid=24','win2') >")
    >>> list(iterlinks(p))
    [Link(url='http://www.example.com/page.html?productid=24', text=None, fragment='', nofollow=False)]

    Dont generate link when target is an anchor
    >>> p = HtmlPage("http://www.example.com", body=u"<html><a href='#section1' >")
    >>> list(iterlinks(p))
    []

    Extract links from <link> tags in page header
    >>> p = HtmlPage("http://example.blogspot.com/", body=u"<html><head><link rel='me' href='http://www.blogger.com/profile/987372' /></head><body>This is my body!</body></html>")
    >>> list(iterlinks(p))
    [Link(url='http://www.blogger.com/profile/987372', text=None, fragment='', nofollow=False)]
    """
    base_href = replace_entities(
        htmlpage.url, encoding=htmlpage.encoding).strip('\\')

    def mklink(url, anchortext=None, nofollow=False):
        url = url.strip()
        fullurl = urljoin(
            base_href, replace_entities(url, encoding=htmlpage.encoding))
        return Link(fullurl.encode(htmlpage.encoding),
                    text=anchortext, nofollow=nofollow)

    # iter to quickly scan only tags
    tag_iter = (t for t in htmlpage.parsed_body if isinstance(t, HtmlTag))

    # parse body
    astart = ahref = None
    nofollow = False
    for nexttag in tag_iter:
        tagname = nexttag.tag
        attributes = nexttag.attributes
        if tagname == 'a' and (nexttag.tag_type == HtmlTagType.CLOSE_TAG or attributes.get('href') \
                    and not attributes.get('href', '').startswith('#')):
            if astart:
                yield mklink(ahref, htmlpage.body[astart:nexttag.start], nofollow)
                astart = ahref = None
                nofollow = False
            href = attributes.get('href')
            if href:
                ahref = href
                astart = nexttag.end
                nofollow = attributes.get('rel') == u'nofollow'
        elif tagname == 'head':
            # scan ahead until end of head section
            for nexttag in tag_iter:
                tagname = nexttag.tag
                if (tagname == 'head' and \
                        nexttag.tag_type == HtmlTagType.CLOSE_TAG) or \
                        tagname == 'body':
                    break
                if tagname == 'base':
                    href = nexttag.attributes.get('href')
                    if href:
                        joined_base = urljoin(htmlpage.url,
                                              href.strip().strip('\\'),
                                              htmlpage.encoding)
                        base_href = replace_entities(
                            joined_base, encoding=htmlpage.encoding)
                elif tagname == 'meta':
                    attrs = nexttag.attributes
                    if attrs.get('http-equiv') == 'refresh':
                        m = _META_REFRESH_CONTENT_RE.search(attrs.get('content', ''))
                        if m:
                            target = m.group('url')
                            if target:
                                yield mklink(target)
                elif tagname == 'link':
                    href = nexttag.attributes.get('href')
                    if href:
                        yield mklink(href)
        elif tagname == 'area':
            href = attributes.get('href')
            if href:
                nofollow = attributes.get('rel') == u'nofollow'
                yield mklink(href, attributes.get('alt', ''), nofollow)
        elif tagname in ('frame', 'iframe'):
            target = attributes.get('src')
            if target:
                yield mklink(target)
        elif 'onclick' in attributes:
            match = _ONCLICK_LINK_RE.search(attributes["onclick"] or "")
            if not match:
                continue
            target = match.group("url")
            nofollow = attributes.get('rel') == u'nofollow'
            yield mklink(target, nofollow=nofollow)

    if astart:
        yield mklink(ahref, htmlpage.body[astart:])





from os.path import join, abspath, dirname, exists
from slybot import __version__
from setuptools import setup, find_packages
from setuptools.command.bdist_egg import bdist_egg
from setuptools.command.sdist import sdist


def build_js():
    root = abspath(dirname(__file__))
    base_path = abspath(join(root, '..', 'splash_utils'))
    if not exists(base_path):
        base_path = abspath(join(root, '..', 'slyd', 'splash_utils'))

    files = ('waitAsync.js', 'perform_actions.js')
    fdata = []
    for fname in files:
        with open(join(base_path, fname)) as f:
            fdata.append(f.read())
    js_file = abspath(join(root, 'slybot', 'splash-script-combined.js'))
    with open(js_file, 'w') as f:
        f.write(';(function(){\n%s\n})();' % '\n'.join(fdata))


class bdist_egg_command(bdist_egg):
    def run(self):
        build_js()
        bdist_egg.run(self)


class sdist_command(sdist):
    def run(self):
        build_js()
        sdist.run(self)



install_requires = ['Scrapy', 'scrapely', 'loginform', 'lxml', 'jsonschema',
                    'dateparser', 'scrapyjs', 'page_finder', 'six']
extras = {
    'tests': ['nose', 'nose-timer'],
    'clustering': ['page_clustering']
}

setup(name='slybot',
      version=__version__,
      license='BSD',
      description='Slybot crawler',
      author='Scrapy project',
      author_email='info@scrapy.org',
      url='http://github.com/scrapinghub/portia',
      packages=find_packages(exclude=('tests', 'tests.*')),
      platforms=['Any'],
      scripts=['bin/slybot', 'bin/portiacrawl'],
      install_requires=install_requires,
      extras_require=extras,
      package_data={'': ['slybot/splash-script-combined.js']},
      include_package_data=True,

      cmdclass={
          'bdist_egg': bdist_egg_command,
          'sdist': sdist_command
      },

      classifiers=[
          'Development Status :: 4 - Beta',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7'
      ])



from scrapy.exporters import CsvItemExporter
from scrapy.conf import settings


class SlybotCSVItemExporter(CsvItemExporter):
    def __init__(self, *args, **kwargs):
        kwargs['fields_to_export'] = settings.getlist('CSV_EXPORT_FIELDS') or None
        super(SlybotCSVItemExporter, self).__init__(*args, **kwargs)



import json

from scrapy.settings import Settings
from os.path import dirname

from scrapy.http.response.html import HtmlResponse
from scrapy.utils.spider import arg_to_iter

from scrapely.htmlpage import HtmlPage

from slybot.spider import IblSpider
from slybot.plugins.scrapely_annotations.builder import (
    apply_annotations, _clean_annotation_data
)


PATH = dirname(__file__)


def make_spider(start_urls=None, sample=None):
    sample = [] if sample is None else arg_to_iter(sample)
    start_urls = [] if start_urls is None else arg_to_iter(start_urls)
    return {'start_urls': start_urls, 'templates': sample}


def open_spec(name):
    use_json = True if name.endswith('.json') else False
    with open('%s/data/templates/%s' % (PATH, name)) as f:
        return json.load(f) if use_json else f.read()


def open_sample_and_page(name):
    sample_spec = open_spec(name)
    annotations = sample_spec['plugins']['annotations-plugin']['extracts']
    annotated = apply_annotations(_clean_annotation_data(annotations),
                                  sample_spec['original_body'])
    url = sample_spec['url']
    return (HtmlPage(url=url, body=annotated),
            HtmlPage(url=url, body=sample_spec['original_body']))


def open_spider_page_and_results(name):
    sample_spec = open_spec(name)
    schemas = sample_spec['schemas']
    results = sample_spec['results']
    page = HtmlResponse('http://url', body=sample_spec['original_body'],
                        encoding='utf-8')
    spider = IblSpider(name, make_spider(sample=sample_spec), schemas, {},
                       Settings())
    return spider, page, results



from scrapy import Request


class FeedGenerator(object):
    def __init__(self, callback):
        self.callback = callback

    def __call__(self, url):
        return Request(url, callback=self.parse_urls)

    def parse_urls(self, response):
        newline_urls = response.text.split('\n')
        urls = [url.replace('\r', '') for url in newline_urls if url]
        for url in urls:
            yield Request(url, callback=self.callback)
